from logging import getLogger
from pathlib import Path
import urllib

import pytest
from requests import PreparedRequest
import requests_mock as rq_mock
from requests_mock import Mocker

from nbgrader.coursedir import CourseDirectory
from .. import Exchange


def parse_body(body: str):
    # https://stackoverflow.com/questions/48018622/how-can-see-the-request-data#51052385
    return dict(urllib.parse.parse_qsl(body))


class TestExchange:
    course_id = 'abc101'
    assignment_id = 'ps1'
    student_id = 'student_1'
    notebook_id = 'p1'
    test_failed = False
    test_completed = False

    def _init_cache_dir(self, tmpdir_factory):
        return Path(tmpdir_factory.mktemp('nbgrader_cache')).absolute()

    def _init_course_dir(self, tmpdir_factory):
        return Path(tmpdir_factory.mktemp(self.course_id)).absolute()

    def _mock_all(self, request: PreparedRequest, content):
        getLogger().fatal(
            'The request "%s" has not been mocked yet.', request.url
        )
        content.status_code = 404
        return ''

    def _new_exchange_object(self, cls, course_id, assignment_id, student_id):
        assert issubclass(cls, Exchange)
        cls.cache = str(self.cache_dir)
        coursedir = CourseDirectory()
        coursedir.root = str(self.course_dir)
        coursedir.course_id = course_id
        coursedir.assignment_id = assignment_id
        obj = cls(coursedir=coursedir)
        obj.username = student_id
        return obj

    @property
    def files_path(self) -> Path:
        return Path(__file__).parent / 'files'

    @pytest.fixture(autouse=True)
    def init(self, requests_mock: Mocker, tmpdir_factory):
        Exchange._ngshare_url = "http://example.com"
        self.base_url = Exchange.ngshare_url.fget(Exchange)
        self.course_dir = self._init_course_dir(tmpdir_factory)
        self.cache_dir = self._init_cache_dir(tmpdir_factory)
        self.requests_mocker = requests_mock
        requests_mock.register_uri(
            rq_mock.ANY, rq_mock.ANY, text=self._mock_all
        )

    def mock_404(self):
        self.requests_mocker.register_uri(
            rq_mock.ANY, rq_mock.ANY, status_code=404
        )

    def mock_unsuccessful(self):
        self.requests_mocker.register_uri(
            rq_mock.ANY,
            rq_mock.ANY,
            json={'success': False, 'message': 'Something happened'},
        )



import os
import sys
import getopt
import requests
import csv
import pwd
import grp
import subprocess
import json
import argparse

# https://www.geeksforgeeks.org/print-colors-python-terminal/
def prRed(skk, exit=True):
    print('\033[91m {}\033[00m'.format(skk))

    if exit:
        sys.exit(-1)


def prGreen(skk):
    print('\033[92m {}\033[00m'.format(skk))


class User:
    def __init__(self, id, first_name, last_name, email):
        self.id = id
        self.first_name = '' if first_name is None else first_name
        self.last_name = '' if last_name is None else last_name
        self.email = '' if email is None else email


def get_username():
    if 'JUPYTERHUB_USER' in os.environ:
        return os.environ['JUPYTERHUB_USER']
    else:
        return os.environ['USER']


def ngshare_url():
    global _ngshare_url
    try:
        return _ngshare_url
    except NameError:
        try:
            from nbgrader.apps import NbGrader

            nbgrader = NbGrader()
            nbgrader.load_config_file()
            exchange = nbgrader.config.ExchangeFactory.exchange()
            _ngshare_url = exchange.ngshare_url
            return _ngshare_url
        except Exception as e:
            prRed(
                "Cannot determine ngshare URL. Please check your nbgrader_config.py!",
                False,
            )
            prRed(e)


def get_header():
    if 'JUPYTERHUB_API_TOKEN' in os.environ:
        return {'Authorization': 'token ' + os.environ['JUPYTERHUB_API_TOKEN']}
    else:
        return None


def check_status_code(response):
    if response.status_code != requests.codes.ok:
        prRed(
            'ngshare returned an invalid status code {}'.format(
                response.status_code
            ),
            False,
        )
        if response.status_code >= 500:
            prRed(
                'ngshare encountered an error. Please contact the maintainers'
            )

        check_message(response)


def check_message(response):

    response = response.json()
    if not response['success']:
        prRed(response['message'])

    return response


def post(url, data):
    header = get_header()

    try:
        response = requests.post(url, data=data, headers=header)
        response.raise_for_status()
    except requests.exceptions.ConnectionError:
        prRed('Could not establish connection to ngshare server')
    except Exception:
        check_status_code(response)

    return check_message(response)


def delete(url, data):
    header = get_header()

    try:
        response = requests.delete(url, data=data, headers=header)
        response.raise_for_status
    except Exception:
        check_status_code(response)

    return check_message(response)


def create_course(course_id, instructors):
    url = '{}/course/{}'.format(ngshare_url(), course_id)
    data = {'user': get_username(), 'instructors': json.dumps(instructors)}

    response = post(url, data)
    prGreen('Successfully created {}'.format(course_id))


def add_student(course_id, student: User, gb):
    # add student to ngshare
    url = '{}/student/{}/{}'.format(ngshare_url(), course_id, student.id)
    data = {
        'user': get_username(),
        'first_name': student.first_name,
        'last_name': student.last_name,
        'email': student.email,
    }

    response = post(url, data)
    prGreen('Successfully added/updated {} on {}'.format(student.id, course_id))

    if gb:
        add_jh_student(student)


def add_jh_student(student: User):
    # add student to nbgrader gradebook
    command = 'nbgrader db student add '

    if len(student.first_name) > 0:
        command += '--first-name {} '.format(student.first_name)
    if len(student.last_name) > 0:
        command += '--last-name {} '.format(student.last_name)
    if len(student.email) > 0:
        command += '--email {} '.format(student.email)

    command += student.id
    os.system(command)


def add_students(course_id, students_csv, gb):
    students = []
    with open(students_csv, 'r') as f:
        csv_reader = csv.reader(f, delimiter=',')
        rows = list(csv_reader)
        if len(rows) == 0:
            prRed('The csv file you entered is empty')

        header = rows[0]

        required_cols = ['student_id', 'first_name', 'last_name', 'email']

        cols_dict = dict()
        for i, col in enumerate(header):
            cols_dict[col] = i

        for col in required_cols:
            if col not in cols_dict:
                prRed('Missing column {} in {}.'.format(col, students_csv))

        for i, row in enumerate(rows[1:]):
            student_dict = {}
            student_id = row[cols_dict['student_id']]
            if len(student_id.replace(' ', '')) == 0:
                prRed('Student ID cannot be empty (row {})'.format(i + 1))
                continue
            first_name = row[cols_dict['first_name']]
            last_name = row[cols_dict['last_name']]
            email = row[cols_dict['email']]

            student_dict['username'] = student_id
            student_dict['first_name'] = first_name
            student_dict['last_name'] = last_name
            student_dict['email'] = email
            students.append(student_dict)

    url = '{}/students/{}'.format(ngshare_url(), course_id)
    data = {'user': get_username(), 'students': json.dumps(students)}

    response = post(url, data)

    if response['success']:
        for i, s in enumerate(response['status']):
            user = s['username']
            if s['success']:
                prGreen('{} was sucessfuly added to {}'.format(user, course_id))
                student = User(
                    user,
                    students[i]['first_name'],
                    students[i]['last_name'],
                    students[i]['email'],
                )
                if gb:
                    add_jh_student(student)
            else:
                prRed(
                    'There was an error adding {} to {}: {}'.format(
                        user, course_id, s['message']
                    ),
                    False,
                )


def remove_jh_student(student_id, force):
    # remove a student from nbgrader gradebook
    command = 'nbgrader db student remove {} '.format(student_id)
    if force:
        command += '--force'
    os.system(command)


def remove_student(course_id, student_id, gb, force):
    if gb:
        remove_jh_student(student_id, force)

    url = '{}/student/{}/{}'.format(ngshare_url(), course_id, student_id)
    data = {'user': get_username()}
    response = delete(url, data)
    prGreen('Successfully deleted {} from {}'.format(student_id, course_id))


def add_instructor(course_id, instructor: User):
    url = '{}/instructor/{}/{}'.format(ngshare_url(), course_id, instructor.id)
    data = {
        'user': get_username(),
        'first_name': instructor.first_name,
        'last_name': instructor.last_name,
        'email': instructor.email,
    }
    print(data)
    response = post(url, data)
    prGreen(
        'Successfully added {} as an instructor to {}'.format(
            instructor.id, course_id
        )
    )


def remove_instructor(course_id, instructor_id):
    url = '{}/instructor/{}/{}'.format(ngshare_url(), course_id, instructor_id)
    data = {'user': get_username()}
    response = delete(url, data)
    prGreen(
        'Successfully deleted instructor {} from {}'.format(
            instructor_id, course_id
        )
    )


def parse_input(argv):
    parser = argparse.ArgumentParser(description='ngshare Course Management')
    parser.add_argument(
        '-c', '--course_id', default=None, help='A unique name for the course'
    )
    parser.add_argument(
        '-s', '--student_id', default=None, help='The ID given to a student'
    )
    parser.add_argument(
        '-i',
        '--instructor_id',
        default=None,
        help='The ID given to an instructor',
    )
    parser.add_argument(
        '-f',
        '--first_name',
        default=None,
        help='First name of the user you are creating',
    )
    parser.add_argument(
        '-l',
        '--last_name',
        default=None,
        help='Last name of the user you are creating',
    )
    parser.add_argument(
        '-e',
        '--email',
        default=None,
        help='Last name of the user you are creating',
    )
    parser.add_argument(
        '--students_csv',
        default=None,
        help='csv file containing a list of students to add. See students.csv as an example.',
    )

    parser.add_argument(
        '--instructors',
        nargs='*',
        default=[],
        help='List of course instructors',
    )

    parser.add_argument(
        'command',
        action='store',
        type=str,
        choices=[
            'create_course',
            'add_student',
            'add_students',
            'remove_student',
            'add_instructor',
            'remove_instructor',
        ],
        help='Command to execute',
    )
    parser.add_argument(
        '--gb',
        action='store_true',
        default=False,
        help='Add student to nbgrader gradebook',
    )

    parser.add_argument(
        '--force',
        action='store_true',
        default=False,
        help='Force gradebook action',
    )

    args = parser.parse_args()

    return args


def execute_command(args):
    command = args.command
    course_id = args.course_id
    student_id = args.student_id
    instructor_id = args.instructor_id
    instructors = args.instructors
    first_name = args.first_name
    last_name = args.last_name
    email = args.email
    students_csv = args.students_csv
    gb = args.gb
    force = args.force

    if command == 'create_course':
        if not course_id:
            prRed(
                'Please specify the course_id for the course with -c or --course_id'
            )
        create_course(course_id, instructors)
    elif command == 'add_student':
        if not course_id:
            prRed(
                'Please specify the course you want to add the student to using -c or --course_id'
            )
        if not student_id:
            prRed(
                'Please specify the student you want to add using -s or --student_id'
            )
        student = User(student_id, first_name, last_name, email)
        add_student(course_id, student, gb)
    elif command == 'add_students':
        if not course_id:
            prRed(
                'Please specify the course you want to add the students to using -c or --course_id'
            )
        if not students_csv:
            prRed(
                'Please enter the path containing the students csv using --students_csv'
            )
        if not os.path.exists(students_csv):
            prRed(
                'The csv file you entered does not exist. Please enter a valid path using --students_csv'
            )
        add_students(course_id, students_csv, gb)
    elif command == 'remove_student':
        if not course_id:
            prRed(
                'Please specify which course you want to remove the student from using -c or --course_id'
            )
        if not student_id:
            prRed(
                'Please specify the student you want to remove from the course using -s or --student_id'
            )
        remove_student(course_id, student_id, gb, force)
    elif command == 'add_instructor':
        if not course_id:
            prRed(
                'Please specify the course you want to add the instructor to using -c or --course_id'
            )
        if not instructor_id:
            prRed(
                'Please specify the instructor you want to add using -i or --instructor_student'
            )
        instructor = User(instructor_id, first_name, last_name, email)
        add_instructor(course_id, instructor)
    elif command == 'remove_instructor':
        if not course_id:
            prRed(
                'Please specify which course you want to remove the instructor from using -c or --course_id'
            )
        if not instructor_id:
            prRed(
                'Please specify the instructor you want to remove using -i or --instructor_id'
            )
        remove_instructor(course_id, instructor_id)

    return True


def main(argv=None):

    argv = argv or sys.argv
    parsed_args = parse_input(argv)
    execute_command(parsed_args)


if __name__ == '__main__':

    sys.exit(main())



import base64
import os
import json

from nbgrader.exchange.abc import ExchangeSubmit as ABCExchangeSubmit
from .exchange import Exchange
from nbgrader.utils import find_all_notebooks, parse_utc


class ExchangeSubmit(Exchange, ABCExchangeSubmit):
    def _get_assignment_notebooks(self, course_id, assignment_id):
        """
        Returns a list of relative paths for all files in the assignment.
        """
        url = '/assignment/{}/{}'.format(course_id, assignment_id)
        params = {'list_only': 'true'}

        response = self.ngshare_api_get(url, params)
        if response is None:
            return None

        return [
            x['path']
            for x in response['files']
            if os.path.splitext(x['path'])[1] == '.ipynb'
        ]

    def init_src(self):
        if self.path_includes_course:
            root = os.path.join(
                self.coursedir.course_id, self.coursedir.assignment_id
            )
            other_path = os.path.join(self.coursedir.course_id, "*")
        else:
            root = self.coursedir.assignment_id
            other_path = "*"
        self.src_path = os.path.abspath(os.path.join(self.assignment_dir, root))
        self.coursedir.assignment_id = os.path.split(self.src_path)[-1]
        if not os.path.isdir(self.src_path):
            self._assignment_not_found(
                self.src_path, os.path.abspath(other_path)
            )

    def init_dest(self):
        if self.coursedir.course_id == '':
            self.fail("No course id specified. Re-run with --course flag.")

        self.cache_path = os.path.join(self.cache, self.coursedir.course_id)
        if self.coursedir.student_id != '*':
            self.fail(
                'Submitting assignments with an explicit student ID is '
                'not possible with ngshare.'
            )
        else:
            student_id = self.username

    def check_filename_diff(self):
        released_notebooks = self._get_assignment_notebooks(
            self.coursedir.course_id, self.coursedir.assignment_id
        )
        if released_notebooks is None:
            self.log.warning('Unable to get list of assignment files.')
            released_notebooks = []
        submitted_notebooks = find_all_notebooks(self.src_path)

        # Look for missing notebooks in submitted notebooks
        missing = False
        release_diff = list()
        for filename in released_notebooks:
            if filename in submitted_notebooks:
                release_diff.append("{}: {}".format(filename, 'FOUND'))
            else:
                missing = True
                release_diff.append("{}: {}".format(filename, 'MISSING'))

        # Look for extra notebooks in submitted notebooks
        extra = False
        submitted_diff = list()
        for filename in submitted_notebooks:
            if filename in released_notebooks:
                submitted_diff.append("{}: {}".format(filename, 'OK'))
            else:
                extra = True
                submitted_diff.append("{}: {}".format(filename, 'EXTRA'))

        if missing or extra:
            diff_msg = "Expected:\n\t{}\nSubmitted:\n\t{}".format(
                '\n\t'.join(release_diff), '\n\t'.join(submitted_diff),
            )
            if missing and self.strict:
                self.fail(
                    "Assignment {} not submitted. "
                    "There are missing notebooks for the submission:\n{}"
                    "".format(self.coursedir.assignment_id, diff_msg)
                )
            else:
                self.log.warning(
                    "Possible missing notebooks and/or extra notebooks "
                    "submitted for assignment {}:\n{}"
                    "".format(self.coursedir.assignment_id, diff_msg)
                )

    def post_submission(self, src_path):
        encoded_dir = self.encode_dir(src_path, ignore=self.ignore_patterns())
        url = '/submission/{}/{}'.format(
            self.coursedir.course_id, self.coursedir.assignment_id
        )

        response = self.ngshare_api_post(url, encoded_dir)
        if response is None:
            return None
        return response['timestamp']

    def copy_files(self):
        self.log.info("Source: {}".format(self.src_path))

        # copy to the real location
        self.check_filename_diff()
        self.timestamp = self.post_submission(self.src_path)
        if self.timestamp is None:
            self.log.error('Failed to submit.')
            return

        # also copy to the cache
        cache_path = os.path.join(
            self.cache_path,
            '{}+{}+{}'.format(
                self.username, self.coursedir.assignment_id, self.timestamp
            ),
        )
        if not os.path.isdir(self.cache_path):
            os.makedirs(self.cache_path)
        self.do_copy(self.src_path, cache_path)
        with open(os.path.join(cache_path, "timestamp.txt"), "w") as fh:
            fh.write(self.timestamp)

        self.log.info(
            "Submitted as: {} {} {}".format(
                self.coursedir.course_id,
                self.coursedir.assignment_id,
                str(self.timestamp),
            )
        )



import base64
import hashlib
from pathlib import Path
import os

import pytest

from .base import TestExchange
from nbgrader.auth import Authenticator
from nbgrader.exchange import ExchangeError
from .. import ExchangeFetchFeedback


class TestExchangeFetchFeedback(TestExchange):

    timestamp = 'some_timestamp'

    def _mock_requests_fetch(self):
        '''
        Mock's ngshare's GET feedback, which responds with the feedback file.
        '''

        url = '{}/feedback/{}/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id, self.student_id
        )
        content = None
        with open(self.files_path / 'feedback.html', 'rb') as feedback:
            content = feedback.read()

        md5 = hashlib.md5()
        md5.update(content)
        checksum = md5.hexdigest()
        content = base64.b64encode(content).decode()
        files = [
            {
                'path': self.notebook_id + '.html',
                'content': content,
                'checksum': checksum,
            }
        ]

        response = {
            'success': True,
            'timestamp': self.timestamp,
            'files': files,
        }
        self.requests_mocker.get(url, json=response)

    def _new_fetch_feedback(
        self,
        course_id=TestExchange.course_id,
        assignment_id=TestExchange.assignment_id,
        student_id=TestExchange.student_id,
    ):

        retvalue = self._new_exchange_object(
            ExchangeFetchFeedback, course_id, assignment_id, student_id
        )

        class DummyAuthenticator(Authenticator):
            def has_access(self, student_id, course_id):
                return True

        retvalue.authenticator = DummyAuthenticator()
        retvalue.assignment_dir = str(self.course_dir.absolute())
        return retvalue

    @pytest.fixture(autouse=True)
    def init_fetch_feedback(self, tmpdir_factory):
        assignment_dir = self.course_dir / self.assignment_id
        os.makedirs(assignment_dir)
        self.fetch_feedback = self._new_fetch_feedback()
        self._mock_requests_fetch()

    def test_404(self):
        self.mock_404()
        self.fetch_feedback.start()

    def test_unsuccessful(self):
        self.mock_unsuccessful()
        self.fetch_feedback.start()

    def test_no_course_id(self):
        self.fetch_feedback.coursedir.course_id = ''
        try:
            self.fetch_feedback.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_bad_student_id(self):
        self.fetch_feedback.coursedir.student_id = '***'
        try:
            self.fetch_feedback.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_fetch(self):

        # set chache folder

        submission_name = '{}+{}+{}'.format(
            self.student_id, self.assignment_id, self.timestamp
        )
        timestamp_path = self.cache_dir / self.course_id / submission_name
        os.makedirs(timestamp_path)

        timestamp_file = timestamp_path / 'timestamp.txt'
        with open(timestamp_file, 'w') as f:
            f.write(self.timestamp)

        self.fetch_feedback.start()

        feedback_path = (
            self.course_dir
            / self.assignment_id
            / 'feedback'
            / self.timestamp
            / (self.notebook_id + '.html')
        )
        assert feedback_path.is_file()
        with open(self.files_path / 'feedback.html', 'rb') as reference_file:
            with open(feedback_path, 'rb') as actual_file:
                assert actual_file.read() == reference_file.read()

    def test_wrong_student_id(self):
        self.fetch_feedback.coursedir.student_id = 'xx+xx'

        with pytest.raises(ExchangeError):
            self.fetch_feedback.start()

    def test_fetch_multiple_feedback(self):
        timestamp1 = 'timestamp1'
        timestamp2 = 'timestamp2'

        self.timestamp = timestamp1
        submission_name = '{}+{}+{}'.format(
            self.student_id, self.assignment_id, self.timestamp
        )
        timestamp_path = self.cache_dir / self.course_id / submission_name
        os.makedirs(timestamp_path)
        timestamp_file = timestamp_path / 'timestamp.txt'
        with open(timestamp_file, 'w') as f:
            f.write(self.timestamp)

        self.fetch_feedback.start()
        self.timestamp = timestamp2
        submission_name = '{}+{}+{}'.format(
            self.student_id, self.assignment_id, self.timestamp
        )
        timestamp_path = self.cache_dir / self.course_id / submission_name

        os.makedirs(timestamp_path)
        timestamp_file = timestamp_path / 'timestamp.txt'
        with open(timestamp_file, 'w') as f:
            f.write(self.timestamp)

        self.fetch_feedback.start()

        feedback_path1 = (
            self.course_dir
            / self.assignment_id
            / 'feedback'
            / timestamp1
            / (self.notebook_id + '.html')
        )
        feedback_path2 = (
            self.course_dir
            / self.assignment_id
            / 'feedback'
            / timestamp2
            / (self.notebook_id + '.html')
        )

        assert feedback_path1.is_file()
        assert feedback_path2.is_file()

    def test_fetch_multiple_courses(self, tmpdir_factory):
        submission_name = '{}+{}+{}'.format(
            self.student_id, self.assignment_id, self.timestamp
        )
        timestamp_path = self.cache_dir / self.course_id / submission_name
        os.makedirs(timestamp_path)
        timestamp_file = timestamp_path / 'timestamp.txt'
        with open(timestamp_file, 'w') as f:
            f.write(self.timestamp)
        self.fetch_feedback.start()
        feedback_path = (
            self.course_dir
            / self.assignment_id
            / 'feedback'
            / self.timestamp
            / (self.notebook_id + '.html')
        )
        assert feedback_path.is_file()

        self.course_id = 'abc102'
        self.course_dir = Path(tmpdir_factory.mktemp(self.course_id))
        assignment_dir = self.course_dir / self.assignment_id
        os.makedirs(assignment_dir)
        self.fetch_feedback = self._new_fetch_feedback(course_id=self.course_id)
        self.timestamp = 'some_other_timestamp'
        submission_name = '{}+{}+{}'.format(
            self.student_id, self.assignment_id, self.timestamp
        )
        timestamp_path = self.cache_dir / self.course_id / submission_name
        os.makedirs(timestamp_path)
        timestamp_file = timestamp_path / 'timestamp.txt'
        with open(timestamp_file, 'w') as f:
            f.write(self.timestamp)

        self._mock_requests_fetch()
        self.fetch_feedback.start()
        feedback_path = (
            self.course_dir
            / self.assignment_id
            / 'feedback'
            / self.timestamp
            / (self.notebook_id + '.html')
        )
        assert feedback_path.is_file()



import os
import glob
import shutil
import re
import hashlib

from nbgrader.exchange.abc import ExchangeList as ABCExchangeList
from .exchange import Exchange


def _checksum(path):
    m = hashlib.md5()
    m.update(open(path, 'rb').read())
    return m.hexdigest()


def _merge_notebooks_feedback(notebook_ids, checksums):
    """
    Returns a list of dictionaries with "notebook_id" and "feedback_checksum".

    ``notebook_ids`` - A list of notebook IDs.
    ``checksum`` - A dictionary mapping notebook IDs to checksums.
    """
    merged = []
    for nb_id in notebook_ids:
        if nb_id not in checksums.keys():
            checksum = None
        else:
            checksum = checksums[nb_id]
        merged.append({'notebook_id': nb_id, 'feedback_checksum': checksum})
    return merged


def _parse_notebook_id(path, extension='.ipynb'):
    """
    Returns the notebook_id from the path. If the path is not a file with the
    extension, returns None.
    """
    split_name = os.path.splitext(os.path.split(path)[1])
    if split_name[1] == extension:
        return split_name[0]
    return None


class ExchangeList(Exchange, ABCExchangeList):
    def _get_assignments(self, course_ids):
        """
        Returns a list of assignments. Each assignment is a dictionary
        containing the course_id and assignment_id.

        ``course_ids`` - A list of course IDs.
        """
        assignments = []
        for course_id in course_ids:
            response = self.ngshare_api_get('/assignments/{}'.format(course_id))
            if response is None:
                self.log.error(
                    'Failed to get assignments from course {}.'.format(
                        course_id
                    )
                )
                continue

            assignments += [
                {'course_id': course_id, 'assignment_id': x}
                for x in response['assignments']
            ]

        return assignments

    def _get_courses(self):
        """
        Returns a list of course_ids.
        """

        response = self.ngshare_api_get('/courses')
        if response is None:
            return None
        return response['courses']

    def _get_feedback_checksums(
        self, course_id, assignment_id, student_id, timestamp
    ):
        """
        Returns the checksums of all feedback files for a specific submission.
        This is a dictionary mapping all notebook_ids to the feedback file's
        checksum.
        """
        url = '/feedback/{}/{}/{}'.format(course_id, assignment_id, student_id)
        params = {'list_only': 'true', 'timestamp': timestamp}

        response = self.ngshare_api_get(url, params)
        if response is None:
            return None

        checksums = {}
        for file_entry in response['files']:
            notebook_id = _parse_notebook_id(file_entry['path'], '.html')
            if notebook_id is not None:
                checksums[notebook_id] = file_entry['checksum']

        return checksums

    def _get_notebooks(self, course_id, assignment_id):
        """
        Returns a list of notebook_ids from the assignment.
        """
        url = '/assignment/{}/{}'.format(course_id, assignment_id)
        params = {'list_only': 'true'}

        response = self.ngshare_api_get(url, params)
        if response is None:
            return None

        return [
            os.path.splitext(os.path.split(x['path'])[1])[0]
            for x in response['files']
        ]

    def _get_submissions(self, assignments, student_id=None):
        """
        Returns a list of submissions. Each submission is a dictionary
        containing the "course_id", "assignment_id", "student_id", "timestamp"
        and a list of "notebooks". Each notebook is a dictionary containing a
        "notebook_id" and "feedback_checksum".

        ``assignments`` - A list of dictionaries containing "course_id" and
        "assignment_id".
        ``student_id`` - Used to specify a specific student's submissions to
        get. If None, submissions from all students are fetched if permitted.
        """
        submissions = []
        for assignment in assignments:
            course_id = assignment['course_id']
            assignment_id = assignment['assignment_id']
            url = '/submissions/{}/{}'.format(course_id, assignment_id)

            if student_id is not None:
                url += '/' + student_id

            response = self.ngshare_api_get(url)
            if response is None:
                self.log.error('Failed to get submisions for assignment {}.')
                continue

            for submission in response['submissions']:
                notebook_ids = self._get_submission_notebooks(
                    course_id,
                    assignment_id,
                    submission['student_id'],
                    submission['timestamp'],
                )
                if notebook_ids is None:
                    self.log.error(
                        'Failed to list notebooks in submission '
                        '{}/{} from student {} (timestamp {})'.format(
                            course_id,
                            assignment_id,
                            submission['student_id'],
                            submission['timestamp'],
                        )
                    )
                    continue
                feedback_checksums = self._get_feedback_checksums(
                    course_id,
                    assignment_id,
                    submission['student_id'],
                    submission['timestamp'],
                )
                if feedback_checksums is None:
                    self.log.error('Failed to check for feedback.')
                    feedback_checksums = []
                notebooks = _merge_notebooks_feedback(
                    notebook_ids, feedback_checksums
                )
                submissions.append(
                    {
                        'course_id': course_id,
                        'assignment_id': assignment_id,
                        'student_id': submission['student_id'],
                        'timestamp': submission['timestamp'],
                        'notebooks': notebooks,
                    }
                )

        return submissions

    def _get_submission_notebooks(
        self, course_id, assignment_id, student_id, timestamp
    ):
        """
        Returns a list of notebook_ids from a submission.
        """
        url = '/submission/{}/{}/{}'.format(
            course_id, assignment_id, student_id
        )
        params = {'list_only': 'true', 'timestamp': timestamp}

        response = self.ngshare_api_get(url, params)
        if response is None:
            return None

        notebooks = []
        for file_entry in response['files']:
            notebook_id = _parse_notebook_id(file_entry['path'], '.ipynb')
            if notebook_id is not None:
                notebooks.append(notebook_id)

        return notebooks

    def _unrelease_assignment(self, course_id, assignment_id):
        """
        Unrelease a released assignment.
        """
        url = '/assignment/{}/{}'.format(course_id, assignment_id)

        return self.ngshare_api_delete(url)

    def init_src(self):
        pass

    def init_dest(self):
        course_id = (
            self.coursedir.course_id if self.coursedir.course_id else '*'
        )
        assignment_id = (
            self.coursedir.assignment_id
            if self.coursedir.assignment_id
            else '*'
        )
        student_id = (
            self.coursedir.student_id if self.coursedir.student_id else '*'
        )

        if course_id == '*':
            courses = self._get_courses()
            if courses is None:
                self.fail('Failed to get courses.')
        else:
            courses = [course_id]
        if assignment_id == '*':
            assignments = self._get_assignments(courses)
        else:
            assignments = [
                {'course_id': course, 'assignment_id': assignment_id}
                for course in courses
            ]

        if self.inbound:
            if student_id == '*':
                student_id = None
            self.assignments = self._get_submissions(assignments, student_id)
        elif self.cached:
            pattern = os.path.join(
                self.cache,
                course_id,
                '{}+{}+*'.format(student_id, assignment_id),
            )
            self.assignments = sorted(glob.glob(pattern))
            if student_id == '*':
                student_id = None
        else:
            self.assignments = assignments

    def parse_assignment(self, assignment):
        if self.inbound:
            return {
                'course_id': assignment['course_id'],
                'student_id': assignment['student_id'],
                'assignment_id': assignment['assignment_id'],
                'timestamp': assignment['timestamp'],
            }
        elif self.cached:
            regexp = r".*/(?P<course_id>.*)/(?P<student_id>.*)\+(?P<assignment_id>.*)\+(?P<timestamp>.*)"
        else:
            return assignment

        m = re.match(regexp, assignment)
        if m is None:
            raise RuntimeError(
                "Could not match '%s' with regexp '%s'", assignment, regexp
            )
        return m.groupdict()

    def format_inbound_assignment(self, info):
        msg = "{course_id} {student_id} {assignment_id} {timestamp}".format(
            **info
        )
        if info['status'] == 'submitted':
            if info['has_local_feedback'] and not info['feedback_updated']:
                msg += " (feedback already fetched)"
            elif info['has_exchange_feedback']:
                msg += " (feedback ready to be fetched)"
            else:
                msg += " (no feedback available)"
        return msg

    def format_outbound_assignment(self, info):
        msg = "{course_id} {assignment_id}".format(**info)
        if os.path.exists(info['assignment_id']):
            msg += " (already downloaded)"
        return msg

    def copy_files(self):
        pass

    def parse_assignments(self):
        if self.coursedir.student_id:
            courses = self.authenticator.get_student_courses(
                self.coursedir.student_id
            )
        else:
            courses = None

        assignments = []
        for assignment in self.assignments:
            info = self.parse_assignment(assignment)
            if courses is not None and info['course_id'] not in courses:
                continue

            if self.path_includes_course:
                assignment_dir = os.path.join(
                    self.assignment_dir,
                    info['course_id'],
                    info['assignment_id'],
                )
            else:
                assignment_dir = os.path.join(
                    self.assignment_dir, info['assignment_id']
                )

            if self.inbound or self.cached:
                info['status'] = 'submitted'
                if self.cached:
                    info['path'] = assignment
            elif os.path.exists(assignment_dir):
                info['status'] = 'fetched'
                info['path'] = os.path.abspath(assignment_dir)
            else:
                info['status'] = 'released'

            if self.remove:
                info['status'] = 'removed'

            if self.cached or info['status'] == 'fetched':
                notebooks = sorted(
                    glob.glob(os.path.join(info['path'], '*.ipynb'))
                )
            elif self.inbound:

                def nb_key(nb):
                    return nb['notebook_id']

                notebooks = sorted(assignment['notebooks'], key=nb_key)
            else:
                notebooks = sorted(
                    self._get_notebooks(
                        info['course_id'], info['assignment_id']
                    )
                )
                if notebooks is None:
                    self.log.error(
                        'Failed to get list of assignment ' 'notebooks.'
                    )
                    notebooks = []

            if not notebooks:
                self.log.warning(
                    'No notebooks found for assignment "{}" in '
                    'course "{}"'.format(
                        info['assignment_id'], info['course_id']
                    )
                )

            if self.cached:
                feedback_checksums = self._get_feedback_checksums(
                    info['course_id'],
                    info['assignment_id'],
                    info['student_id'],
                    info['timestamp'],
                )
                if feedback_checksums is None:
                    feedback_checksums = {}

            info['notebooks'] = []
            for notebook in notebooks:
                if self.cached:
                    nb_info = {
                        'notebook_id': os.path.splitext(
                            os.path.split(notebook)[1]
                        )[0],
                        'path': os.path.abspath(notebook),
                    }
                elif self.inbound:
                    nb_info = {'notebook_id': notebook['notebook_id']}
                elif info['status'] == 'fetched':
                    nb_info = {
                        'notebook_id': notebook,
                        'path': os.path.abspath(notebook),
                    }
                else:
                    nb_info = {'notebook_id': notebook}
                if info['status'] != 'submitted':
                    info['notebooks'].append(nb_info)
                    continue

                nb_info['has_local_feedback'] = False
                nb_info['has_exchange_feedback'] = False
                nb_info['local_feedback_path'] = None
                nb_info['feedback_updated'] = False

                # Check whether feedback has been fetched already.
                local_feedback_dir = os.path.join(
                    assignment_dir, 'feedback', info['timestamp']
                )
                local_feedback_path = os.path.join(
                    local_feedback_dir,
                    '{0}.html'.format(nb_info['notebook_id']),
                )
                has_local_feedback = os.path.isfile(local_feedback_path)
                if has_local_feedback:
                    local_feedback_checksum = _checksum(local_feedback_path)
                else:
                    local_feedback_checksum = None

                # Also look to see if there is feedback available to fetch.
                if self.cached:
                    has_exchange_feedback = False
                    exchange_feedback_checksum = None
                    notebook_id = nb_info['notebook_id']
                    if notebook_id in feedback_checksums.keys():
                        has_exchange_feedback = True
                        exchange_feedback_checksum = feedback_checksums[
                            notebook_id
                        ]
                else:  # self.inbound
                    has_exchange_feedback = (
                        notebook['feedback_checksum'] is not None
                        and notebook['feedback_checksum'] != ''
                    )
                    if has_exchange_feedback:
                        exchange_feedback_checksum = notebook[
                            'feedback_checksum'
                        ]
                    else:
                        exchange_feedback_checksum = None

                nb_info['has_local_feedback'] = has_local_feedback
                nb_info['has_exchange_feedback'] = has_exchange_feedback
                if has_local_feedback:
                    nb_info['local_feedback_path'] = local_feedback_path
                if has_local_feedback and has_exchange_feedback:
                    nb_info['feedback_updated'] = (
                        exchange_feedback_checksum != local_feedback_checksum
                    )
                info['notebooks'].append(nb_info)

            if info['status'] == 'submitted':
                if info['notebooks']:
                    has_local_feedback = all(
                        [nb['has_local_feedback'] for nb in info['notebooks']]
                    )
                    has_exchange_feedback = all(
                        [
                            nb['has_exchange_feedback']
                            for nb in info['notebooks']
                        ]
                    )
                    feedback_updated = any(
                        [nb['feedback_updated'] for nb in info['notebooks']]
                    )
                else:
                    has_local_feedback = False
                    has_exchange_feedback = False
                    feedback_updated = False

                info['has_local_feedback'] = has_local_feedback
                info['has_exchange_feedback'] = has_exchange_feedback
                info['feedback_updated'] = feedback_updated
                if has_local_feedback:
                    info['local_feedback_path'] = os.path.join(
                        assignment_dir, 'feedback', info['timestamp']
                    )
                else:
                    info['local_feedback_path'] = None

            assignments.append(info)

        # partition the assignments into groups for course/student/assignment
        if self.inbound or self.cached:
            _get_key = lambda info: (
                info['course_id'],
                info['student_id'],
                info['assignment_id'],
            )
            _match_key = lambda info, key: (
                info['course_id'] == key[0]
                and info['student_id'] == key[1]
                and info['assignment_id'] == key[2]
            )
            assignment_keys = sorted(
                list(set([_get_key(info) for info in assignments]))
            )
            assignment_submissions = []
            for key in assignment_keys:
                submissions = [x for x in assignments if _match_key(x, key)]
                submissions = sorted(submissions, key=lambda x: x['timestamp'])
                info = {
                    'course_id': key[0],
                    'student_id': key[1],
                    'assignment_id': key[2],
                    'status': submissions[0]['status'],
                    'submissions': submissions,
                }
                assignment_submissions.append(info)
            assignments = assignment_submissions

        return assignments

    def list_files(self):
        """List files."""
        assignments = self.parse_assignments()

        if self.inbound or self.cached:
            self.log.info("Submitted assignments:")
            for assignment in assignments:
                for info in assignment['submissions']:
                    self.log.info(self.format_inbound_assignment(info))
        else:
            self.log.info("Released assignments:")
            for info in assignments:
                self.log.info(self.format_outbound_assignment(info))

        return assignments

    def remove_files(self):
        """List and remove files."""
        assignments = self.parse_assignments()

        if self.inbound or self.cached:
            self.log.info("Removing submitted assignments:")
            for assignment in assignments:
                for info in assignment['submissions']:
                    self.log.info(self.format_inbound_assignment(info))
        else:
            self.log.info("Removing released assignments:")
            for info in assignments:
                self.log.info(self.format_outbound_assignment(info))

        if self.cached:
            for assignment in self.assignments:
                shutil.rmtree(assignment)
        elif self.inbound:
            self.log.warning(
                'ngshare does not support removing submissions.'
            )  # TODO
        else:
            for assignment in self.assignments:
                retvalue = self._unrelease_assignment(
                    assignment['course_id'], assignment['assignment_id']
                )
                if retvalue is None:
                    self.log.error(
                        'Failed to remove assignment {}/{}.'.format(
                            assignment['course_id'], assignment['assignment_id']
                        )
                    )

        return assignments



import base64
import hashlib
import logging
import os
from pathlib import Path
import re
import shutil

from _pytest.logging import LogCaptureFixture
from _pytest.tmpdir import TempdirFactory
import pytest
from requests import PreparedRequest
from textwrap import dedent

from .base import parse_body, TestExchange
from nbgrader.auth import Authenticator
from nbgrader.exchange import ExchangeError
from .. import ExchangeList


class TestExchangeList(TestExchange):
    course_id2 = 'xyz200'
    assignment_id2 = 'ps2'
    timestamp1 = '2000-02-03 12:34:56.789012 UTC'
    timestamp2 = '2001-02-03 12:34:56.789012 UTC'

    def _delete_assignment(self, request: PreparedRequest, context):
        if not self.is_instructor:
            return {'success': False, 'message': 'Permission denied'}
        self.test_completed = True
        return {'success': True}

    def _feedback_content(self):
        reference_file = self.files_path / 'feedback.html'
        with open(reference_file, 'rb') as file:
            return file.read()

    def _fetch(self, course_dir, assignment_id=TestExchange.assignment_id):
        notebook_path = (
            course_dir / assignment_id / (self.notebook_id + '.ipynb')
        )
        notebook_path.parent.mkdir()
        shutil.copyfile(self.files_path / 'test.ipynb', notebook_path)
        pass

    def _fetch_feedback(self, course_dir, course_id, assignment_id, timestamp):
        feedback_dir = course_dir / assignment_id / 'feedback' / timestamp
        feedback_dir.mkdir(parents=True)
        html_path = feedback_dir / (self.notebook_id + '.html')
        timestamp_path = feedback_dir / 'timestamp.txt'
        shutil.copyfile(self.files_path / 'feedback.html', html_path)
        with open(timestamp_path, 'w') as timestamp_file:
            timestamp_file.write(timestamp)

    def _get_assignment(self, request: PreparedRequest, context):
        content = self._notebook_content()
        md5 = hashlib.md5()
        md5.update(content)
        checksum = md5.hexdigest()
        files = [
            {
                'path': self.notebook_id + '.ipynb',
                'content': base64.b64encode(content).decode(),
                'checksum': checksum,
            }
        ]
        return {'success': True, 'files': files}

    def _get_assignments(self, request: PreparedRequest, context):
        assignments = [self.assignment_id, self.assignment_id2]
        return {
            'success': True,
            'assignments': assignments[: self.num_assignments],
        }

    def _get_courses(self, request: PreparedRequest, context):
        courses = [self.course_id, self.course_id2]
        return {'success': True, 'courses': courses[: self.num_courses]}

    def _get_feedback(self, request: PreparedRequest, context):
        pattern = re.compile(r'\?[^#]*')
        matches = pattern.findall(request.url)
        if len(matches) == 0 or 'timestamp' not in parse_body(matches[0]):
            print(request)
            return {'success': False, 'message': 'Please supply timestamp'}
        timestamp = parse_body(matches[0])['timestamp']
        if timestamp not in [self.timestamp1, self.timestamp2]:
            return {'success': False, 'message': 'Submission not found'}
        files = []
        if (timestamp == self.timestamp1 and self.num_feedback > 0) or (
            timestamp == self.timestamp2 and self.num_feedback > 1
        ):
            content = self._feedback_content()
            md5 = hashlib.md5()
            md5.update(content)
            checksum = md5.hexdigest()
            files = [
                {
                    'path': self.notebook_id + '.html',
                    'content': base64.b64encode(content).decode(),
                    'checksum': checksum,
                }
            ]
        return {'success': True, 'timestamp': timestamp, 'files': files}

    def _get_student_submissions(self, request: PreparedRequest, context):
        submissions = [
            {'student_id': self.student_id, 'timestamp': self.timestamp1},
            {'student_id': self.student_id, 'timestamp': self.timestamp2},
        ]
        return {
            'success': True,
            'submissions': submissions[: self.num_submissions],
        }

    def _get_submission(self, request: PreparedRequest, context):
        if not self.is_instructor:
            return {'success': False, 'message': 'Permission denied'}
        return self._get_assignment(request, context)

    def _get_submissions(self, request: PreparedRequest, context):
        return self._get_student_submissions(request, context)

    def _mock_requests_list(self):
        """
        Mocks ngshare's GET courses, GET assignments, GET assignment, DELETE
        assignment, GET submissions GET student's submissions, GET submission,
        and GET feedback.
        """
        url = '{}/courses'.format(self.base_url)
        self.requests_mocker.get(url, json=self._get_courses)

        url = '{}/assignments/{}'.format(self.base_url, self.course_id)
        self.requests_mocker.get(url, json=self._get_assignments)
        url = '{}/assignments/{}'.format(self.base_url, self.course_id2)
        self.requests_mocker.get(url, json=self._get_assignments)

        url = '{}/assignment/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id
        )
        self.requests_mocker.get(url, json=self._get_assignment)
        url = '{}/assignment/{}/{}'.format(
            self.base_url, self.course_id2, self.assignment_id
        )
        self.requests_mocker.get(url, json=self._get_assignment)
        url = '{}/assignment/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id2
        )
        self.requests_mocker.get(url, json=self._get_assignment)
        url = '{}/assignment/{}/{}'.format(
            self.base_url, self.course_id2, self.assignment_id2
        )
        self.requests_mocker.get(url, json=self._get_assignment)

        url = '{}/assignment/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id
        )
        self.requests_mocker.delete(url, json=self._delete_assignment)

        url = '{}/submissions/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id
        )
        self.requests_mocker.get(url, json=self._get_submissions)

        url = '{}/submissions/{}/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id, self.student_id
        )
        self.requests_mocker.get(url, json=self._get_student_submissions)

        url = '{}/submission/{}/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id, self.student_id
        )
        self.requests_mocker.get(url, json=self._get_submission)

        url = '{}/feedback/{}/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id, self.student_id
        )
        self.requests_mocker.get(url, json=self._get_feedback)
        url = '{}/feedback/{}/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id2, self.student_id
        )
        self.requests_mocker.get(url, json=self._get_feedback)

    def _new_list(
        self,
        course_id=TestExchange.course_id,
        assignment_id=TestExchange.assignment_id,
        student_id=TestExchange.student_id,
    ):
        retvalue = self._new_exchange_object(
            ExchangeList, course_id, assignment_id, student_id
        )

        class DummyAuthenticator(Authenticator):
            def has_access(self, student_id, course_id):
                return True

        retvalue.authenticator = DummyAuthenticator()
        retvalue.assignment_dir = str(self.course_dir.absolute())
        return retvalue

    def _notebook_content(self):
        reference_file = self.files_path / 'test.ipynb'
        with open(reference_file, 'rb') as file:
            return file.read()

    def _read_log(self):
        log_records = [
            '[{}] {}\n'.format(x.levelname, x.getMessage())
            for x in self.caplog.get_records('call')
        ]
        self.caplog.clear()
        return ''.join(log_records)

    def _submit(
        self,
        course_id=TestExchange.course_id,
        assignment_id=TestExchange.assignment_id,
        timestamp=timestamp1,
    ):
        assignment_filename = '{}+{}+{}'.format(
            self.student_id, assignment_id, timestamp
        )
        notebook_path = (
            self.cache_dir
            / course_id
            / assignment_filename
            / (self.notebook_id + '.ipynb')
        )
        timestamp_path = notebook_path.parent / 'timestamp.txt'
        notebook_path.parent.mkdir(parents=True)
        shutil.copyfile(self.files_path / 'test.ipynb', notebook_path)
        with open(timestamp_path, 'w') as timestamp_file:
            timestamp_file.write(timestamp)

    @pytest.fixture(autouse=True)
    def init_submit(
        self, caplog: LogCaptureFixture, tmpdir_factory: TempdirFactory
    ):
        self.caplog = caplog
        self.course_dir2 = Path(
            tmpdir_factory.mktemp(self.course_id2)
        ).absolute()
        self._mock_requests_list()
        self.list = self._new_list()
        os.chdir(self.course_dir)
        self.caplog.set_level(logging.INFO)
        self.list.coursedir.course_id = '*'
        self.list.coursedir.assignment_id = '*'
        self.num_courses = 1
        self.num_assignments = 0
        self.num_submissions = 0
        self.num_feedback = 0
        self.is_instructor = True

    def test_404(self):
        self.mock_404()
        try:
            self.list.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_unsuccessful(self):
        self.mock_unsuccessful()
        try:
            self.list.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_list_released_2x1_course1(self):
        self.num_courses = 2
        self.num_assignments = 1
        self.list.coursedir.course_id = self.course_id
        self.list.start()
        output = self._read_log()
        assert (
            output
            == dedent(
                """
            [INFO] Released assignments:
            [INFO] {} {}
            """
            )
            .lstrip()
            .format(self.course_id, self.assignment_id)
        )

    def test_list_released_2x1_course2(self):
        self.num_courses = 2
        self.num_assignments = 1
        self.list.coursedir.course_id = self.course_id2
        self.list.start()
        output = self._read_log()
        assert (
            output
            == dedent(
                """
            [INFO] Released assignments:
            [INFO] {} {}
            """
            )
            .lstrip()
            .format(self.course_id2, self.assignment_id)
        )

    def test_list_released_2x1(self):
        self.num_courses = 2
        self.num_assignments = 1
        self.list.coursedir.course_id = '*'
        self.list.start()
        output = self._read_log()
        assert (
            output
            == dedent(
                """
            [INFO] Released assignments:
            [INFO] {} {}
            [INFO] {} {}
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.assignment_id,
                self.course_id2,
                self.assignment_id,
            )
        )

    def test_list_released_2x2_assignment1(self):
        self.num_courses = 2
        self.num_assignments = 2
        self.list.coursedir.assignment_id = self.assignment_id
        self.list.start()
        output = self._read_log()
        assert (
            output
            == dedent(
                """
            [INFO] Released assignments:
            [INFO] {} {}
            [INFO] {} {}
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.assignment_id,
                self.course_id2,
                self.assignment_id,
            )
        )

    def test_list_released_2x2_assignment2(self):
        self.num_courses = 2
        self.num_assignments = 2
        self.list.coursedir.assignment_id = self.assignment_id2
        self.list.start()
        output = self._read_log()
        assert (
            output
            == dedent(
                """
            [INFO] Released assignments:
            [INFO] {} {}
            [INFO] {} {}
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.assignment_id2,
                self.course_id2,
                self.assignment_id2,
            )
        )

    def test_list_released_2x2(self):
        self.num_courses = 2
        self.num_assignments = 2
        self.list.start()
        output = self._read_log()
        assert (
            output
            == dedent(
                """
            [INFO] Released assignments:
            [INFO] {} {}
            [INFO] {} {}
            [INFO] {} {}
            [INFO] {} {}
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.assignment_id,
                self.course_id,
                self.assignment_id2,
                self.course_id2,
                self.assignment_id,
                self.course_id2,
                self.assignment_id2,
            )
        )

    def test_list_fetched(self):
        self.num_assignments = 2
        self._fetch(self.course_dir)
        self.list.start()
        output = self._read_log()
        assert (
            output
            == dedent(
                """
            [INFO] Released assignments:
            [INFO] {} {} (already downloaded)
            [INFO] {} {}
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.assignment_id,
                self.course_id,
                self.assignment_id2,
            )
        )

    def test_list_remove_outbound(self):
        self.num_assignments = 2
        self.list.coursedir.assignment_id = self.assignment_id
        self.list.remove = True
        self.list.start()
        assert not self.test_failed
        assert self.test_completed

    def test_list_inbound_0(self):
        self.num_assignments = 1
        self.num_submissions = 0
        self.list.coursedir.assignment_id = self.assignment_id
        self.list.inbound = True
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            """
            ).lstrip()
        )

    def test_list_inbound_1(self):
        self.num_assignments = 1
        self.num_submissions = 1
        self.list.coursedir.assignment_id = self.assignment_id
        self.list.inbound = True
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (no feedback available)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
            )
        )

    def test_list_inbound_2(self):
        self.num_assignments = 1
        self.num_submissions = 2
        self.list.coursedir.assignment_id = self.assignment_id
        self.list.inbound = True
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (no feedback available)
            [INFO] {} {} {} {} (no feedback available)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp2,
            )
        )

    def test_list_cached_0(self):
        self.num_assignments = 1
        self.is_instructor = False
        self.list.cached = True
        self.list.coursedir.assignment_id = self.assignment_id
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            """
            ).lstrip()
        )

    def test_list_cached_1(self):
        self.num_assignments = 1
        self._submit()
        self.is_instructor = False
        self.list.cached = True
        self.list.coursedir.assignment_id = self.assignment_id
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (no feedback available)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
            )
        )

    def test_list_cached_2(self):
        self.num_assignments = 1
        self._submit()
        self._submit(timestamp=self.timestamp2)
        self.is_instructor = False
        self.list.cached = True
        self.list.coursedir.assignment_id = self.assignment_id
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (no feedback available)
            [INFO] {} {} {} {} (no feedback available)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp2,
            )
        )

    def test_list_remove_cached(self):
        self._submit()
        self._submit(
            assignment_id=self.assignment_id2, timestamp=self.timestamp2
        )
        self.is_instructor = False
        self.list.cached = True
        self.list.remove = True
        self.list.coursedir.assignment_id = self.assignment_id
        self.list.start()
        self._read_log()
        self.list.remove = False
        self.list.coursedir.assignment_id = '*'
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (no feedback available)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id2,
                self.timestamp2,
            )
        )

    def test_list_cached_and_inbound(self):
        self.is_instructor = False
        self.list.cached = True
        self.list.inbound = True
        with pytest.raises(ExchangeError):
            self.list.start()

    def test_list_feedback_inbound_ready1(self):
        self.num_assignments = 1
        self.num_submissions = 2
        self.num_feedback = 1
        self.list.inbound = True
        self.list.coursedir.assignment_id = self.assignment_id
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (feedback ready to be fetched)
            [INFO] {} {} {} {} (no feedback available)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp2,
            )
        )

    def test_list_feedback_inbound_fetched1(self):
        self.num_assignments = 1
        self.num_submissions = 2
        self.num_feedback = 1
        self.list.inbound = True
        self.list.coursedir.assignment_id = self.assignment_id
        self._fetch_feedback(
            self.course_dir, self.course_id, self.assignment_id, self.timestamp1
        )
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (feedback already fetched)
            [INFO] {} {} {} {} (no feedback available)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp2,
            )
        )

    def test_list_feedback_inbound_modified1(self):
        self.num_assignments = 1
        self.num_submissions = 2
        self.num_feedback = 1
        self.list.inbound = True
        self.list.coursedir.assignment_id = self.assignment_id
        self._fetch_feedback(
            self.course_dir, self.course_id, self.assignment_id, self.timestamp1
        )
        feedback_path = (
            self.course_dir
            / self.assignment_id
            / 'feedback'
            / self.timestamp1
            / (self.notebook_id + '.html')
        )
        # This part is different from the original test because the fetched
        # feedback is modified instead of the outbound one, but the results
        # should be the same.
        with open(feedback_path, 'a') as fetched_file:
            fetched_file.write('blahblahblah')
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (feedback ready to be fetched)
            [INFO] {} {} {} {} (no feedback available)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp2,
            )
        )

    def test_list_feedback_inbound_modified1_ready1(self):
        self.num_assignments = 1
        self.num_submissions = 2
        self.num_feedback = 1
        self.list.inbound = True
        self.list.coursedir.assignment_id = self.assignment_id
        self._fetch_feedback(
            self.course_dir, self.course_id, self.assignment_id, self.timestamp1
        )
        feedback_path = (
            self.course_dir
            / self.assignment_id
            / 'feedback'
            / self.timestamp1
            / (self.notebook_id + '.html')
        )
        with open(feedback_path, 'a') as fetched_file:
            fetched_file.write('blahblahblah')
        self.num_feedback = 2
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (feedback ready to be fetched)
            [INFO] {} {} {} {} (feedback ready to be fetched)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp2,
            )
        )

    def test_list_feedback_inbound_fetched2(self):
        self.num_assignments = 1
        self.num_submissions = 2
        self.num_feedback = 1
        self.list.inbound = True
        self.list.coursedir.assignment_id = self.assignment_id
        self._fetch_feedback(
            self.course_dir, self.course_id, self.assignment_id, self.timestamp1
        )
        self._fetch_feedback(
            self.course_dir, self.course_id, self.assignment_id, self.timestamp2
        )
        self.num_feedback = 2
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (feedback already fetched)
            [INFO] {} {} {} {} (feedback already fetched)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp2,
            )
        )

    def test_list_feedback_cached_ready1(self):
        self.num_assignments = 1
        self.num_submissions = 2
        self.num_feedback = 1
        self._submit()
        self._submit(timestamp=self.timestamp2)
        self.is_instructor = False
        self.list.cached = True
        self.list.coursedir.assignment_id = self.assignment_id
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (feedback ready to be fetched)
            [INFO] {} {} {} {} (no feedback available)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp2,
            )
        )

    def test_list_feedback_cached_fetched1(self):
        self.num_assignments = 1
        self.num_submissions = 2
        self.num_feedback = 1
        self._submit()
        self._submit(timestamp=self.timestamp2)
        self._fetch_feedback(
            self.course_dir, self.course_id, self.assignment_id, self.timestamp1
        )
        self.is_instructor = False
        self.list.cached = True
        self.list.coursedir.assignment_id = self.assignment_id
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (feedback already fetched)
            [INFO] {} {} {} {} (no feedback available)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp2,
            )
        )

    def test_list_feedback_cached_modified1(self):
        self.num_assignments = 1
        self.num_submissions = 2
        self.num_feedback = 1
        self._submit()
        self._submit(timestamp=self.timestamp2)
        self._fetch_feedback(
            self.course_dir, self.course_id, self.assignment_id, self.timestamp1
        )
        self.is_instructor = False
        self.list.cached = True
        self.list.coursedir.assignment_id = self.assignment_id
        feedback_path = (
            self.course_dir
            / self.assignment_id
            / 'feedback'
            / self.timestamp1
            / (self.notebook_id + '.html')
        )
        # This part is different from the original test because the fetched
        # feedback is modified instead of the outbound one, but the results
        # should be the same.
        with open(feedback_path, 'a') as fetched_file:
            fetched_file.write('blahblahblah')
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (feedback ready to be fetched)
            [INFO] {} {} {} {} (no feedback available)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp2,
            )
        )

    def test_list_feedback_cached_ready2(self):
        self.num_assignments = 1
        self.num_submissions = 2
        self.num_feedback = 1
        self._submit()
        self._submit(timestamp=self.timestamp2)
        self._fetch_feedback(
            self.course_dir, self.course_id, self.assignment_id, self.timestamp1
        )
        self.is_instructor = False
        self.list.cached = True
        self.list.coursedir.assignment_id = self.assignment_id
        feedback_path = (
            self.course_dir
            / self.assignment_id
            / 'feedback'
            / self.timestamp1
            / (self.notebook_id + '.html')
        )
        with open(feedback_path, 'a') as fetched_file:
            fetched_file.write('blahblahblah')
        self.num_feedback = 2
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (feedback ready to be fetched)
            [INFO] {} {} {} {} (feedback ready to be fetched)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp2,
            )
        )

    def test_list_feedback_cached_fetched2(self):
        self.num_assignments = 1
        self.num_submissions = 2
        self.num_feedback = 1
        self._submit()
        self._submit(timestamp=self.timestamp2)
        self._fetch_feedback(
            self.course_dir, self.course_id, self.assignment_id, self.timestamp1
        )
        self._fetch_feedback(
            self.course_dir, self.course_id, self.assignment_id, self.timestamp2
        )
        self.is_instructor = False
        self.list.cached = True
        self.list.coursedir.assignment_id = self.assignment_id
        self.num_feedback = 2
        self.list.start()
        assert (
            self._read_log()
            == dedent(
                """
            [INFO] Submitted assignments:
            [INFO] {} {} {} {} (feedback already fetched)
            [INFO] {} {} {} {} (feedback already fetched)
            """
            )
            .lstrip()
            .format(
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp1,
                self.course_id,
                self.student_id,
                self.assignment_id,
                self.timestamp2,
            )
        )



#!/usr/bin/python
import os
import shutil
from stat import (
    S_IRUSR,
    S_IWUSR,
    S_IXUSR,
    S_IRGRP,
    S_IWGRP,
    S_IXGRP,
    S_IROTH,
    S_IWOTH,
    S_IXOTH,
    S_ISGID,
    ST_MODE,
)

from traitlets import Bool

from nbgrader.exchange.abc import (
    ExchangeReleaseAssignment as ABCExchangeReleaseAssignment,
)
from .exchange import Exchange


class ExchangeReleaseAssignment(Exchange, ABCExchangeReleaseAssignment):

    force = Bool(
        False, help='Force overwrite existing files in the exchange.'
    ).tag(config=True)

    def _load_config(self, cfg, **kwargs):
        if 'ExchangeRelease' in cfg:
            self.log.warning(
                'Use ExchangeReleaseAssignment in config, not ExchangeRelease. Outdated config:\n%s',
                '\n'.join(
                    'ExchangeRelease.{key} = {value!r}'.format(
                        key=key, value=svalue
                    )
                    for (key, value) in cfg.ExchangeRelease.items()
                ),
            )

            cfg.ExchangeReleaseAssignment.merge(cfg.ExchangeRelease)
            del cfg.ExchangeRelease

        super(ExchangeReleaseAssignment, self)._load_config(cfg, **kwargs)

    def ensure_root(self):
        pass

    def init_src(self):
        self.src_path = self.coursedir.format_path(
            self.coursedir.release_directory, '.', self.coursedir.assignment_id
        )

        if not os.path.isdir(self.src_path):
            source = self.coursedir.format_path(
                self.coursedir.source_directory,
                '.',
                self.coursedir.assignment_id,
            )
            if os.path.isdir(source):

                # Looks like the instructor forgot to assign
                self.fail(
                    "Assignment found in '{}' but not '{}', run `nbgrader generate_assignment` first.".format(
                        source, self.src_path
                    )
                )
            else:
                self._assignment_not_found(
                    self.src_path,
                    self.coursedir.format_path(
                        self.coursedir.release_directory, '.', '*'
                    ),
                )

    def init_dest(self):
        if self.coursedir.course_id == '':
            self.fail("No course id specified. Re-run with --course flag.")
        self.dest_path = '/assignment/{}/{}'.format(
            self.coursedir.course_id, self.coursedir.assignment_id
        )

    def assignment_exists(self):
        url = '/assignments/{}'.format(self.coursedir.course_id)
        response = self.ngshare_api_get(url)

        if response is None:
            self.log.error(
                'An error occurred while trying to check if the assignment exists {}.'.format(
                    self.coursedir.course_id
                )
            )
            return True

        if self.coursedir.assignment_id in response['assignments']:
            if self.force:
                self.log.info(
                    "Overwriting files: {} {}".format(
                        self.coursedir.course_id, self.coursedir.assignment_id
                    )
                )
                delete_url = '/assignment/{}/{}'.format(
                    self.coursedir.course_id, self.coursedir.assignment_id
                )
                response = self.ngshare_api_delete(delete_url)
                if response is None:
                    self.fail(
                        'An error occurred while trying to delete {}'.format(
                            self.coursedir.assignment_id
                        )
                    )
                    return True
            else:
                self.fail(
                    'Destination already exists, add --force to overwrite: {} {}'.format(
                        self.coursedir.course_id, self.coursedir.assignment_id
                    )
                )
                return True

        return False

    def copy_files(self):
        if not self.assignment_exists():
            self.log.info('Encoding assignment')
            data = self.encode_dir(self.src_path)
            response = self.ngshare_api_post(self.dest_path, data)
            if response is None:
                self.log.warning(
                    'An error occurred while trying to release {}'.format(
                        self.coursedir.assignment_id
                    )
                )
            else:
                self.log.info(
                    'Successfully released {}'.format(
                        self.coursedir.assignment_id
                    )
                )



import os
import shutil
import sys
from collections import defaultdict
import base64

from nbgrader.exchange.abc import ExchangeCollect as ABCExchangeCollect
from .exchange import Exchange

from nbgrader.utils import parse_utc

# pwd is for matching unix names with student ide, so we shouldn't import it on
# windows machines
if sys.platform != 'win32':
    import pwd
else:
    pwd = None


def groupby(l, key=lambda x: x):
    d = defaultdict(list)
    for item in l:
        d[key(item)].append(item)
    return d


class ExchangeCollect(Exchange, ABCExchangeCollect):
    def _get_submission(self, course_id, assignment_id, student_id):
        """
        Returns the student's submission. A submission is a dictionary
        containing a "timestamp" and "files" list. Each file in the list is a
        dictionary containing the "path" relative to the assignment root and the
        "content" as an ASCII representation of the base64 encoded bytes.
        """

        response = self.ngshare_api_get(
            "/submission/{}/{}/{}".format(course_id, assignment_id, student_id)
        )
        if response is None:
            self.log.error('An error occurred downloading a submission.')
            return None

        timestamp = response['timestamp']
        files = response['files']
        files.append(
            {
                'path': 'timestamp.txt',
                'content': base64.encodebytes(timestamp.encode()).decode(),
            }
        )
        return {'timestamp': timestamp, 'files': files}

    def _get_submission_list(self, course_id, assignment_id):
        """
        Returns a list of submission entries. Each entry is a dictionary
        containing the "student_id" and "timestamp".
        """
        response = self.ngshare_api_get(
            '/submissions/{}/{}'.format(course_id, assignment_id)
        )
        if response is None:
            return None

        return [
            {
                'student_id': x['student_id'],
                'timestamp': parse_utc(x['timestamp']),
            }
            for x in response['submissions']
        ]

    def _sort_by_timestamp(self, records):
        return sorted(records, key=lambda item: item['timestamp'], reverse=True)

    def init_src(self):
        if self.coursedir.course_id == '':
            self.fail("No course id specified. Re-run with --course flag.")

        records = self._get_submission_list(
            self.coursedir.course_id, self.coursedir.assignment_id
        )
        if records is None:
            self.fail('Failed to list submissions.')
        usergroups = groupby(records, lambda item: item['student_id'])
        self.src_records = [
            self._sort_by_timestamp(v)[0] for v in usergroups.values()
        ]

    def init_dest(self):
        pass

    def copy_files(self):
        if len(self.src_records) == 0:
            self.log.warning(
                "No submissions of '{}' for course '{}' to collect".format(
                    self.coursedir.assignment_id, self.coursedir.course_id
                )
            )
        else:
            self.log.info(
                "Processing {} submissions of '{}' for course '{}'".format(
                    len(self.src_records),
                    self.coursedir.assignment_id,
                    self.coursedir.course_id,
                )
            )

        for rec in self.src_records:
            student_id = rec['student_id']

            dest_path = self.coursedir.format_path(
                self.coursedir.submitted_directory,
                student_id,
                self.coursedir.assignment_id,
            )
            if not os.path.exists(os.path.dirname(dest_path)):
                os.makedirs(os.path.dirname(dest_path))

            copy = False
            updating = False
            if os.path.isdir(dest_path):
                existing_timestamp = self.coursedir.get_existing_timestamp(
                    dest_path
                )
                new_timestamp = rec['timestamp']
                if self.update and (
                    existing_timestamp is None
                    or new_timestamp > existing_timestamp
                ):
                    copy = True
                    updating = True
            else:
                copy = True

            if copy:
                if updating:
                    self.log.info(
                        "Updating submission: {} {}".format(
                            student_id, self.coursedir.assignment_id
                        )
                    )
                    shutil.rmtree(dest_path)
                else:
                    self.log.info(
                        "Collecting submission: {} {}".format(
                            student_id, self.coursedir.assignment_id
                        )
                    )
                submission = self._get_submission(
                    self.coursedir.course_id,
                    self.coursedir.assignment_id,
                    student_id,
                )
                if submission is None:
                    continue
                self.do_copy(submission['files'], dest_path)
            else:
                if self.update:
                    self.log.info(
                        "No newer submission to collect: {} {}".format(
                            student_id, self.coursedir.assignment_id
                        )
                    )
                else:
                    self.log.info(
                        "Submission already exists, use --update to update: {} {}".format(
                            student_id, self.coursedir.assignment_id
                        )
                    )

    def do_copy(self, src, dest):
        """
        Repurposed version of Exchange.do_copy.
        """
        self.decode_dir(src, dest)



import base64
from json import loads
import hashlib
from logging import getLogger
import os
from pathlib import Path
import re
import shutil

import pytest
from requests import PreparedRequest
import requests_mock as rq_mock

from .base import parse_body, TestExchange
from nbgrader.exchange import ExchangeError
from .. import ExchangeSubmit


def get_files_path() -> Path:
    return Path(__file__).parent / 'files'


class TestExchangeSubmit(TestExchange):
    timestamp = 'some timestamp'

    def _mock_request_assignment(self):
        '''
        Mocks ngshare's GET assignment with list_only, which responds with the
        assignment.
        '''
        url = '{}/assignment/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id
        )
        pattern = re.compile(r'^{}\?.*list_only=true'.format(url))
        content = None
        with open(self.files_path / 'test.ipynb', 'rb') as notebook:
            content = notebook.read()
        md5 = hashlib.md5()
        md5.update(content)
        checksum = md5.hexdigest()
        files = [{'path': self.notebook_id + '.ipynb', 'checksum': checksum}]
        response = {'success': True, 'files': files}
        self.requests_mocker.get(pattern, json=response)

    def _mock_requests_submit(self, extra=False):
        ''' Mock's ngshare's POST submission, which verifies the request. '''
        url = '{}/submission/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id
        )
        if extra:
            self.requests_mocker.post(url, json=self._post_submission_extra)
        else:
            self.requests_mocker.post(url, json=self._post_submission)

        self._mock_request_assignment()

    def _mock_requests_submit_2(self):
        '''
        Mocks ngshare's POST submission, which verifies the request for the
        second course.
        '''
        url = '{}/submission/{}/{}'.format(
            self.base_url, self.course_id2, self.assignment_id
        )
        self.requests_mocker.post(url, json=self._post_submission)

        self._mock_request_assignment()

    def _mock_requests_submit_size(self):
        '''
        Mocks ngshare's POST submission, which verifies the request has all and
        only the files which are within the file size limit.
        '''
        url = '{}/submission/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id
        )
        self.requests_mocker.post(url, json=self._post_submission_size)

        self._mock_request_assignment()

    def _new_submit(
        self,
        course_id=TestExchange.course_id,
        assignment_id=TestExchange.assignment_id,
        student_id=TestExchange.student_id,
    ):
        return self._new_exchange_object(
            ExchangeSubmit, course_id, assignment_id, student_id
        )

    def _post_submission(self, request: PreparedRequest, context):
        request = parse_body(request.body)
        try:
            files = loads(request['files'])
            assert len(files) == 1
            notebook_name = self.notebook_id + '.ipynb'
            assert files[0]['path'] == notebook_name
            actual_content = base64.b64decode(files[0]['content'].encode())
            reference_file = self.files_path / 'test.ipynb'
            with open(reference_file, 'rb') as expected_content:
                assert actual_content == expected_content.read()
        except Exception as e:
            self.test_failed = True
            getLogger().error(e)
        self.test_completed = True
        return {'success': True, 'timestamp': self.timestamp}

    def _post_submission_extra(self, request: PreparedRequest, context):
        request = parse_body(request.body)
        try:
            files = loads(request['files'])
            assert len(files) == 2
            assert files[0]['path'] != files[1]['path']
            notebook_name1 = self.notebook_id + '.ipynb'
            notebook_name2 = self.notebook_id2 + '.ipynb'
            reference_file = self.files_path / 'test.ipynb'
            expected_content = None
            with open(reference_file, 'rb') as reference_file:
                expected_content = reference_file.read()
            for file_entry in files:
                assert (
                    file_entry['path'] == notebook_name1
                    or file_entry['path'] == notebook_name2
                )
                actual_content = base64.b64decode(files[0]['content'].encode())
                assert actual_content == expected_content
        except Exception as e:
            self.test_failed = True
            getLogger().error(e)
        self.test_completed = True
        return {'success': True, 'timestamp': self.timestamp}

    def _post_submission_size(self, request: PreparedRequest, context):
        request = parse_body(request.body)
        try:
            files = loads(request['files'])
            paths = [x['path'] for x in files]
            assert 'small_file' in paths
            assert 'big_file' not in paths
        except Exception as e:
            self.test_failed = True
            getLogger().error(e)
        self.test_completed = True
        return {'success': True, 'timestamp': self.timestamp}

    def _prepare_submission(
        self,
        course_dir,
        assignment_id=TestExchange.assignment_id,
        notebook_id=TestExchange.notebook_id,
    ):
        course_dir = Path(course_dir).absolute()
        assignment_dir = course_dir / assignment_id
        os.makedirs(assignment_dir)
        shutil.copyfile(
            self.files_path / 'test.ipynb',
            assignment_dir / (notebook_id + '.ipynb'),
        )

    @pytest.fixture(autouse=True)
    def init_submit(self):
        self._prepare_submission(self.course_dir)
        self.submit = self._new_submit()
        os.chdir(self.course_dir)

    def test_404(self):
        self.mock_404()
        try:
            self.submit.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_unsuccessful(self):
        self.mock_unsuccessful()
        try:
            self.submit.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_no_course_id(self):
        """Does submitting without a course id thrown an error?"""
        self._mock_requests_submit()
        self.submit.coursedir.course_id = ''
        with pytest.raises(ExchangeError):
            self.submit.start()

    def test_submit(self):
        self._mock_requests_submit()
        self.submit.start()
        assert not self.test_failed
        assert self.test_completed

        # Verify cache.
        (cache_filename,) = os.listdir(self.cache_dir / self.course_id)
        (
            cache_username,
            cache_assignment,
            cache_timestamp1,
        ) = cache_filename.split("+")[:3]
        assert cache_username == self.student_id
        assert cache_assignment == self.assignment_id
        assert cache_timestamp1 == self.timestamp
        assert Path(
            self.cache_dir
            / self.course_id
            / cache_filename
            / (self.notebook_id + '.ipynb')
        ).is_file()
        assert Path(
            self.cache_dir / self.course_id / cache_filename / 'timestamp.txt'
        ).is_file()
        with open(
            self.cache_dir / self.course_id / cache_filename / 'timestamp.txt',
            'r',
        ) as fh:
            assert fh.read() == cache_timestamp1

        # Submit again.
        self.test_failed = False
        self.test_completed = False
        self.timestamp += '_1'
        self.submit.start()
        assert not self.test_failed
        assert self.test_completed

        # Verify cache.
        assert len(os.listdir(self.cache_dir / self.course_id)) == 2
        cache_filename = sorted(os.listdir(self.cache_dir / self.course_id))[1]
        (
            cache_username,
            cache_assignment,
            cache_timestamp2,
        ) = cache_filename.split("+")[:3]
        assert cache_username == self.student_id
        assert cache_assignment == self.assignment_id
        assert cache_timestamp2 == self.timestamp
        assert Path(
            self.cache_dir
            / self.course_id
            / cache_filename
            / (self.notebook_id + '.ipynb')
        ).is_file()
        assert Path(
            self.cache_dir / self.course_id / cache_filename / 'timestamp.txt'
        ).is_file()
        with open(
            self.cache_dir / self.course_id / cache_filename / 'timestamp.txt',
            'r',
        ) as fh:
            assert fh.read() == cache_timestamp2

    def test_submit_extra(self):
        # Add extra notebook.
        self._mock_requests_submit(extra=True)
        self.notebook_id2 = 'p2'
        shutil.copyfile(
            get_files_path() / 'test.ipynb',
            self.course_dir
            / self.assignment_id
            / (self.notebook_id2 + '.ipynb'),
        )
        self.submit.start()
        assert not self.test_failed
        assert self.test_completed

    def test_submit_extra_strict(self):
        # Add extra notebook and enable strict flag.
        self._mock_requests_submit(extra=True)
        self.notebook_id2 = 'p2'
        shutil.copyfile(
            get_files_path() / 'test.ipynb',
            self.course_dir
            / self.assignment_id
            / (self.notebook_id2 + '.ipynb'),
        )
        self.submit.strict = True
        self.submit.start()
        assert not self.test_failed
        assert self.test_completed

    def test_submit_missing(self):
        # Missing notebook.
        self._mock_requests_submit()
        self.notebook_id2 = 'p2'
        shutil.move(
            self.course_dir
            / self.assignment_id
            / (self.notebook_id + '.ipynb'),
            self.course_dir
            / self.assignment_id
            / (self.notebook_id2 + '.ipynb'),
        )
        self.submit.start()

    def test_submit_missing_strict(self):
        # Missing notebook and enable strict flag.
        self._mock_requests_submit()
        self.notebook_id2 = 'p2'
        shutil.move(
            self.course_dir
            / self.assignment_id
            / (self.notebook_id + '.ipynb'),
            self.course_dir
            / self.assignment_id
            / (self.notebook_id2 + '.ipynb'),
        )
        self.submit.strict = True
        with pytest.raises(ExchangeError):
            self.submit.start()

    def test_submit_multiple_courses(self, tmpdir_factory):
        self._mock_requests_submit()
        self.course_id2 = 'abc102'
        course_dir2 = Path(tmpdir_factory.mktemp(self.course_id2)).absolute()
        self._prepare_submission(course_dir2)

        # Submit to first course.
        self.submit.start()
        assert not self.test_failed
        assert self.test_completed
        assert len(os.listdir(self.cache_dir / self.course_id)) == 1
        assert not Path(self.cache_dir / self.course_id2).exists()

        # Submit to second course.
        self.requests_mocker.register_uri(
            rq_mock.ANY, rq_mock.ANY, text=self._mock_all
        )
        self._mock_requests_submit_2()
        self.course_dir = course_dir2
        self.submit = self._new_submit(course_id=self.course_id2)
        os.chdir(self.course_dir)
        self.submit.start()
        assert not self.test_failed
        assert self.test_completed
        assert len(os.listdir(self.cache_dir / self.course_id)) == 1
        assert len(os.listdir(self.cache_dir / self.course_id2)) == 1

    def test_submit_exclude(self):
        # Create a file that should be ignored.
        self._mock_requests_submit()
        (self.course_dir / self.assignment_id / 'foo.pyc').touch()
        self.submit.start()
        assert not self.test_failed
        assert self.test_completed

    def test_submit_include(self):
        # Create a file that should be ignored.
        self._mock_requests_submit()
        (self.course_dir / self.assignment_id / 'foo.txt').touch()
        self.submit.coursedir.include = ['*.ipynb']
        self.submit.start()
        assert not self.test_failed
        assert self.test_completed

    def test_submit_file_size(self):
        # Create two files around a 2 KB size limit.
        self._mock_requests_submit_size()
        small_file = self.course_dir / self.assignment_id / 'small_file'
        big_file = self.course_dir / self.assignment_id / 'big_file'
        small_file.touch()
        big_file.touch()
        with open(small_file, 'w') as small, open(big_file, 'w') as big:
            small.write('x' * 2000)
            big.write('x' * 2001)
        self.submit.coursedir.max_file_size = 2
        self.submit.start()
        assert not self.test_failed
        assert self.test_completed



import base64
import os

import pytest
from requests import PreparedRequest

from .base import parse_body, TestExchange
from nbgrader.exchange import ExchangeError
from .. import ExchangeCollect
from nbgrader.utils import parse_utc


class TestExchangeCollect(TestExchange):
    def _get_submission(self, request: PreparedRequest, context):
        request = parse_body(request.body)
        timestamp = None
        try:
            if 'timestamp' in request:
                timestamp = request['timestamp']
            else:
                timestamp = (
                    str(2000 + self.num_submissions) + self.timestamp_template
                )
            time = parse_utc(timestamp)
            assert time.year > 2000 and time.year < 2001 + self.num_submissions
            assert timestamp == str(time.year) + self.timestamp_template
        except Exception:
            return {'success': False, 'message': 'Submission not found'}
        content = base64.b64encode(self._notebook_content()).decode()
        files = [{'path': self.notebook_id + '.ipynb', 'content': content}]
        return {'success': True, 'timestamp': timestamp, 'files': files}

    def _get_submissions(self, request: PreparedRequest, context):
        submissions = []
        for i in range(self.num_submissions):
            timestamp = str(2001 + i) + self.timestamp_template
            submissions.append(
                {'student_id': self.student_id, 'timestamp': timestamp}
            )
        return {'success': True, 'submissions': submissions}

    def _mock_requests_collect(self):
        """
        Mock's ngshare's GET submissions, which responds with the submissions,
        and GET submission, which responds with the submission.
        """
        url = '{}/submissions/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id
        )
        self.requests_mocker.get(url, json=self._get_submissions)

        url = '{}/submission/{}/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id, self.student_id
        )
        self.requests_mocker.get(url, json=self._get_submission)

    def _mock_requests_subdir(self, subdirectory, subdirectory_file):
        """
        Mock's ngshare's GET submissions, which responds with the submission,
        and GET submission, which responds with the submission with a
        subdirectory.
        """
        url = '{}/submissions/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id
        )
        timestamp = '2000' + self.timestamp_template
        submissions = [{'student_id': self.student_id, 'timestamp': timestamp}]
        response = {'success': True, 'submissions': submissions}
        self.requests_mocker.get(url, json=response)

        url = '{}/submission/{}/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id, self.student_id
        )
        content1 = base64.b64encode(self._notebook_content()).decode()
        content2 = base64.b64encode(''.encode()).decode()
        path2 = '{}/{}'.format(subdirectory, subdirectory_file)
        files = [
            {'path': self.notebook_id + '.ipynb', 'content': content1},
            {'path': path2, 'content': content2},
        ]
        response = {'success': True, 'timestamp': timestamp, 'files': files}
        self.requests_mocker.get(url, json=response)

    def _new_collect(
        self,
        course_id=TestExchange.course_id,
        assignment_id=TestExchange.assignment_id,
        student_id=TestExchange.student_id,
    ):
        return self._new_exchange_object(
            ExchangeCollect, course_id, assignment_id, student_id
        )

    def _notebook_content(self):
        reference_file = self.files_path / 'test.ipynb'
        with open(reference_file, 'rb') as file:
            return file.read()

    @pytest.fixture(autouse=True)
    def init_collect(self):
        self.collect = self._new_collect()
        os.chdir(self.course_dir)

    def submission_dir(self):
        return (
            self.course_dir / 'submitted' / self.student_id / self.assignment_id
        )

    @property
    def timestamp_template(self):
        return '-12-21 12:34:56.789012 UTC'

    def test_404(self):
        self.mock_404()
        try:
            self.collect.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_unsuccessful(self):
        self.mock_unsuccessful()
        try:
            self.collect.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_no_course_id(self):
        """Does collecting without a course id throw an error?"""
        self._mock_requests_collect()
        with pytest.raises(ExchangeError):
            self.collect.start()

    def test_collect_0(self):
        self.num_submissions = 0
        self._mock_requests_collect()
        self.collect.start()
        assert not (self.course_dir / 'submitted').is_dir()

    def test_collect_1(self):
        self.num_submissions = 1
        self._mock_requests_collect()
        self.collect.start()
        notebook = self.submission_dir() / (self.notebook_id + '.ipynb')
        timestamp_path = self.submission_dir() / 'timestamp.txt'
        assert notebook.is_file()
        assert timestamp_path.is_file()
        with open(notebook, 'rb') as notebook_file, open(
            timestamp_path, 'r'
        ) as timestamp_file:
            assert notebook_file.read() == self._notebook_content()
            assert timestamp_file.read() == '2001' + self.timestamp_template

    def test_collect_1_twice(self):
        self.num_submissions = 1
        self._mock_requests_collect()
        self.collect.start()
        self.collect.start()
        timestamp_path = self.submission_dir() / 'timestamp.txt'
        with open(timestamp_path, 'r') as timestamp_file:
            assert timestamp_file.read() == '2001' + self.timestamp_template

    def test_collect_no_update(self):
        self.num_submissions = 1
        self._mock_requests_collect()
        self.collect.start()
        self.num_submissions = 2
        self.collect.start()
        timestamp_path = self.submission_dir() / 'timestamp.txt'
        with open(timestamp_path, 'r') as timestamp_file:
            assert timestamp_file.read() == '2001' + self.timestamp_template

    def test_collect_update(self):
        self.num_submissions = 1
        self._mock_requests_collect()
        self.collect.start()
        self.num_submissions = 2
        self.collect.update = True
        self.collect.start()
        timestamp_path = self.submission_dir() / 'timestamp.txt'
        with open(timestamp_path, 'r') as timestamp_file:
            assert timestamp_file.read() == '2002' + self.timestamp_template

    def test_collect_subdirectories(self):
        subdir = 'foo'
        subfile = 'temp.txt'
        self.num_submissions = 1
        self._mock_requests_subdir(subdir, subfile)
        self.collect.start()
        assert (
            self.course_dir
            / 'submitted'
            / self.student_id
            / self.assignment_id
            / subdir
            / subfile
        ).is_file()



#!/usr/bin/python
import os
import shutil
import glob

from nbgrader.exchange.abc import (
    ExchangeFetchFeedback as ABCExchangeFetchFeedback,
)
from .exchange import Exchange

from nbgrader.utils import (
    check_mode,
    notebook_hash,
    make_unique_key,
    get_username,
)
from nbgrader.utils import parse_utc


class ExchangeFetchFeedback(Exchange, ABCExchangeFetchFeedback):
    def init_src(self):
        if self.coursedir.course_id == '':
            self.fail('No course id specified. Re-run with --course flag.')

        if self.coursedir.student_id != '*':

            # An explicit student id has been specified on the command line; we use it as student_id
            if (
                '*' in self.coursedir.student_id
                or '+' in self.coursedir.student_id
            ):
                self.fail(
                    "The student ID should contain no '*' nor '+'; got {}".format(
                        self.coursedir.student_id
                    )
                )
            student_id = self.coursedir.student_id
        else:
            student_id = get_username()
        self.cache_path = os.path.join(self.cache, self.coursedir.course_id)
        assignment_id = (
            self.coursedir.assignment_id
            if self.coursedir.assignment_id
            else '*'
        )
        pattern = os.path.join(self.cache_path, '*+{}+*'.format(assignment_id))
        self.log.debug(
            'Looking for submissions with pattern: {}'.format(pattern)
        )

        self.src_path = '/feedback/{}/{}/{}'.format(
            self.coursedir.course_id, assignment_id, self.username
        )

        self.timestamps = []
        submissions = [os.path.split(x)[-1] for x in glob.glob(pattern)]
        for submission in submissions:
            (_, assignment_id, timestamp) = submission.split('/')[-1].split('+')
            self.timestamps.append(timestamp)

    def init_dest(self):
        if self.path_includes_course:
            root = os.path.join(
                self.coursedir.course_id, self.coursedir.assignment_id
            )
        else:
            root = self.coursedir.assignment_id
        self.dest_path = os.path.abspath(
            os.path.join(self.assignment_dir, root, 'feedback')
        )

        # check if feedback folder exists
        if not os.path.exists(self.dest_path):
            os.mkdir(self.dest_path)

    def copy_files(self):
        self.log.info('Fetching feedback from server')
        if len(self.timestamps) == 0:
            self.log.warning(
                'No feedback available to fetch for your submissions'
            )

        for timestamp in self.timestamps:
            params = {'timestamp': timestamp, 'list_only': 'false'}
            response = self.ngshare_api_get(self.src_path, params=params)
            if response is None:
                self.log.warning(
                    'An error occurred while trying to fetch feedback for {}'.format(
                        self.coursedir.assignment_id
                    )
                )
                return
            try:
                dest_with_timestamp = os.path.join(
                    self.dest_path, str(timestamp)
                )
                self.decode_dir(response['files'], dest_with_timestamp)
                self.log.info(
                    'Successfully decoded feedback for {} saved to {}'.format(
                        self.coursedir.assignment_id, dest_with_timestamp
                    )
                )
            except:
                self.log.warning(
                    'Could not decode feedback for timestamp {}'.format(
                        str(timestamp)
                    )
                )



import base64
import os

import pytest
import re
from logging import getLogger
import requests
from requests import PreparedRequest
import requests_mock as rq_mock
from requests_mock import Mocker
import urllib
import tempfile
from .. import course_management as cm
from io import StringIO


class Command:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


def remove_color(s):
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    result = ansi_escape.sub('', s)
    return result


def parse_body(body: str):
    # https://stackoverflow.com/questions/48018622/how-can-see-the-request-data#51052385
    return dict(urllib.parse.parse_qsl(body))


NGSHARE_URL = 'http://127.0.0.1:12121/api'
global _ngshare_url
cm._ngshare_url = NGSHARE_URL


class TestCourseManagement:
    course_id = 'math101'
    instructor = (
        os.environ['JUPYTERHUB_USER']
        if 'JUPYTERHUB_USER' in os.environ
        else os.environ['USER']
    )
    instructors = ['mi1', 'mi2']
    student_id = 'ms'
    instructor_id = 'mi1'

    course_created = False

    @pytest.fixture(autouse=True)
    def init(self, requests_mock: Mocker):
        self.requests_mocker = requests_mock
        requests_mock.register_uri(
            rq_mock.ANY, rq_mock.ANY, text=self._mock_all
        )

    def form_command(
        self,
        command,
        course_id=None,
        student_id=None,
        instructor_id=None,
        instructors=[],
        first_name=None,
        last_name=None,
        email=None,
        students_csv=None,
        gb=False,
        force=False,
    ):
        return Command(
            command=command,
            course_id=course_id,
            student_id=student_id,
            instructor_id=instructor_id,
            instructors=instructors,
            first_name=first_name,
            last_name=last_name,
            email=email,
            students_csv=students_csv,
            gb=gb,
            force=force,
        )

    def _mock_all(self, request: PreparedRequest, content):
        getLogger().fatal(
            'The request \'%s\' has not been mocked yet.', request.url
        )
        content.status_code = 404
        return ''

    def _get_user_info(self, request: PreparedRequest, context):
        request = parse_body(request.body)
        if 'first_name' not in request:
            return {'success': False, 'message': 'Please supply first name'}
        elif 'last_name' not in request:
            return {'success': False, 'message': 'Please supply last name'}
        elif 'email' not in request:
            return {'success': False, 'message': 'Please supply email'}
        elif request['user'] != self.instructor:
            return {'success': False, 'message': 'Permission denied'}
        else:
            return {'success': True}

    def _get_students_info(self, request: PreparedRequest, context):
        request = parse_body(request.body)
        students = eval(request['students'])

        if (
            'sid1' == students[0]['username']
            and 'sid2' == students[1]['username']
        ):
            return {
                'success': True,
                'status': [
                    {'username': 'sid1', 'success': True},
                    {'username': 'sid2', 'success': True},
                ],
            }
        else:
            return {'success': False, 'message': 'wrong students passed in'}

    def _get_user(self, request: PreparedRequest, context):
        request = parse_body(request.body)
        response = {'success': False, 'message': 'Permission denied'}
        if self.instructor in request['user']:
            response = {'success': True}
        return response

    def _get_instructors_info(self, request: PreparedRequest, context):
        request = parse_body(request.body)
        response = {'success': False, 'message': 'Some error occurred'}
        if not self.course_created:
            if 'instructors' in request:
                instructors = eval(request['instructors'])
                if (
                    self.instructors[0] == instructors[0]
                    and self.instructors[1] == instructors[1]
                ):
                    response = {'success': True}
                    self.course_created = True
        else:
            response = {'success': False, 'message': 'Course already exists'}

        return response

    def _mock_create_course(self):
        url = '{}/course/{}'.format(NGSHARE_URL, self.course_id)
        self.requests_mocker.post(url, json=self._get_instructors_info)

    def _mock_add_student(self):
        url = '{}/student/{}/{}'.format(
            NGSHARE_URL, self.course_id, self.student_id
        )
        self.requests_mocker.post(url, json=self._get_user_info)

    def _mock_add_students(self):
        url = '{}/students/{}'.format(
            NGSHARE_URL, self.course_id, self.student_id
        )
        self.requests_mocker.post(url, json=self._get_students_info)

    def _mock_add_instructor(self):
        url = '{}/instructor/{}/{}'.format(
            NGSHARE_URL, self.course_id, self.instructor_id
        )
        self.requests_mocker.post(url, json=self._get_user_info)

    def _mock_remove_student(self):
        url = '{}/student/{}/{}'.format(
            NGSHARE_URL, self.course_id, self.student_id
        )
        self.requests_mocker.delete(url, json=self._get_user)

    def _mock_remove_instructor(self):
        url = '{}/instructor/{}/{}'.format(
            NGSHARE_URL, self.course_id, self.instructor_id
        )
        self.requests_mocker.delete(url, json=self._get_user)

    def test_crete_course(self, capsys):
        self._mock_create_course()
        cmd = self.form_command(
            'create_course',
            course_id=self.course_id,
            instructors=self.instructors,
        )
        cm.execute_command(cmd)
        out, err = capsys.readouterr()
        out = remove_color(out)
        assert ' Successfully created {}\n'.format(self.course_id) in out

        # test missing course id
        with pytest.raises(SystemExit) as se:
            cmd = self.form_command('create_course')
            cm.execute_command(cmd)
        assert se.type == SystemExit
        assert se.value.code == -1

        # try to create course again
        self._mock_create_course()
        with pytest.raises(SystemExit) as se:
            cmd = self.form_command('create_course', course_id=self.course_id)
            cm.execute_command(cmd)
        out, err = capsys.readouterr()
        assert ' Course already exists' in out
        assert se.type == SystemExit
        assert se.value.code == -1

    def test_add_student(self, capsys):
        # test missing course id
        with pytest.raises(SystemExit) as se:
            cmd = self.form_command('add_student')
            cm.execute_command(cmd)
        assert se.type == SystemExit
        assert se.value.code == -1

        # test missing student id
        with pytest.raises(SystemExit) as se:
            cmd = self.form_command('add_student', course_id=self.course_id)
            cm.execute_command(cmd)
        assert se.type == SystemExit
        assert se.value.code == -1

        self._mock_add_student()
        cmd = self.form_command(
            'add_student',
            course_id=self.course_id,
            student_id=self.student_id,
            first_name='jane',
            last_name='doe',
            email='jd@mail.com',
        )
        cm.execute_command(cmd)
        out, err = capsys.readouterr()
        assert 'Successfully added/updated {}'.format(self.student_id) in out

    def test_add_students(self, capsys, tmp_path):
        self._mock_add_students()
        # test no course id
        with pytest.raises(SystemExit) as se:
            cmd = self.form_command('add_students')
            cm.execute_command(cmd)
        assert se.type == SystemExit
        assert se.value.code == -1

        # test no file
        with pytest.raises(SystemExit) as se:
            cmd = self.form_command('add_students', course_id=self.course_id)
            cm.execute_command(cmd)
        assert se.type == SystemExit
        assert se.value.code == -1

        # test no non existing file
        with pytest.raises(SystemExit) as se:
            cmd = self.form_command(
                'add_students', course_id=self.course_id, students_csv='dne'
            )
            cm.execute_command(cmd)

        assert se.type == SystemExit
        assert se.value.code == -1
        out, err = capsys.readouterr()
        assert 'The csv file you entered does not exist' in out

        with tempfile.NamedTemporaryFile() as f:
            f.writelines(
                [
                    b"student_id,first_name,last_name,email\n",
                    b"sid1,jane,doe,jd@mail.com\n",
                    b"sid2,john,perez,jp@mail.com\n",
                ]
            )
            f.flush()
            cmd = self.form_command(
                'add_students', course_id=self.course_id, students_csv=f.name,
            )
            cm.execute_command(cmd)
        out, err = capsys.readouterr()
        assert 'sid1 was sucessfuly added to math101' in out
        assert 'sid2 was sucessfuly added to math101' in out

    def test_add_instructor(self, capsys):
        self._mock_add_instructor()

        # test no course id
        with pytest.raises(SystemExit) as se:
            cmd = self.form_command('add_instructor')
            cm.execute_command(cmd)
        assert se.type == SystemExit
        assert se.value.code == -1

        # test no instructor id
        with pytest.raises(SystemExit) as se:
            cmd = self.form_command('add_instructor', course_id=self.course_id)
            cm.execute_command(cmd)
        assert se.type == SystemExit
        assert se.value.code == -1

        # test valid
        cmd = self.form_command(
            'add_instructor',
            course_id=self.course_id,
            instructor_id=self.instructor_id,
            first_name='john',
            last_name='doe',
            email='jd@mail.com',
        )
        cm.execute_command(cmd)
        out, err = capsys.readouterr()
        assert (
            'Successfully added {} as an instructor to {}'.format(
                self.instructor_id, self.course_id
            )
            in out
        )

    def test_remove_student(self, capsys):
        self._mock_remove_student()

        # test missing course id
        with pytest.raises(SystemExit) as se:
            cmd = self.form_command('remove_student')
            cm.execute_command(cmd)
        assert se.type == SystemExit
        assert se.value.code == -1

        # test missing student id
        with pytest.raises(SystemExit) as se:
            cmd = self.form_command('remove_student', course_id=self.course_id)
            cm.execute_command(cmd)
        assert se.type == SystemExit
        assert se.value.code == -1

        # test valid
        cmd = self.form_command(
            'remove_student',
            course_id=self.course_id,
            student_id=self.student_id,
        )
        cm.execute_command(cmd)
        out, err = capsys.readouterr()
        assert (
            'Successfully deleted {} from {}'.format(
                self.student_id, self.course_id
            )
            in out
        )

    def test_remove_instructor(self, capsys):
        self._mock_remove_instructor()

        # test missing course id
        with pytest.raises(SystemExit) as se:
            cmd = self.form_command('remove_instructor')
            cm.execute_command(cmd)
        assert se.type == SystemExit
        assert se.value.code == -1

        # test missing student id
        with pytest.raises(SystemExit) as se:
            cmd = self.form_command(
                'remove_instructor', course_id=self.course_id
            )
            cm.execute_command(cmd)
        assert se.type == SystemExit
        assert se.value.code == -1

        # test valid
        cmd = self.form_command(
            'remove_instructor',
            course_id=self.course_id,
            instructor_id=self.instructor_id,
        )
        cm.execute_command(cmd)
        out, err = capsys.readouterr()
        assert (
            'Successfully deleted instructor {} from {}'.format(
                self.instructor_id, self.course_id
            )
            in out
        )

    def test_add_students_parsing(self, capsys):
        # test empty file
        with tempfile.NamedTemporaryFile() as f:
            with pytest.raises(SystemExit) as se:
                cm.add_students(self.course_id, f.name, False)
            assert se.type == SystemExit
            assert se.value.code == -1
            out, err = capsys.readouterr()
            assert 'The csv file you entered is empty' in out

        # test missing a column
        with tempfile.NamedTemporaryFile() as f:
            f.write(b'first_name,last_name,email')
            f.flush()

            with pytest.raises(SystemExit) as se:
                cm.add_students(self.course_id, f.name, False)
            assert se.type == SystemExit
            assert se.value.code == -1
            out, err = capsys.readouterr()
            assert (
                'Missing column {} in {}.'.format('student_id', f.name) in out
            )

        with tempfile.NamedTemporaryFile() as f:
            f.write(b'student_id,first_name,last_name,email\n')
            f.write(b',jane,doe,jd@mail.com')
            f.flush()

            with pytest.raises(SystemExit) as se:
                cm.add_students(self.course_id, f.name, False)
            assert se.type == SystemExit
            assert se.value.code == -1
            out, err = capsys.readouterr()
            assert 'Student ID cannot be empty (row 1)' in out



import base64
from json import loads
from logging import getLogger
from pathlib import Path
import os
import shutil

import pytest
from requests import PreparedRequest

from .base import parse_body, TestExchange
from nbgrader.exchange import ExchangeError
from .. import ExchangeReleaseAssignment


class TestExchangeReleaseAssignment(TestExchange):
    def _delete_assignment(self, request: PreparedRequest, context):
        if not self.released:
            return {'success': False, 'message': 'Assignment not found'}
        self.released = False
        return {'success': True}

    def _get_assignments(self, request: PreparedRequest, context):
        if self.released:
            return {'success': True, 'assignments': [self.assignment_id]}
        return {'success': True, 'assignments': []}

    def _mock_requests_release(self):
        '''
        Mocks ngshare's GET assignments, which responds with no assignments,
        and POST assignment, which verifies the request.
        '''
        url = '{}/assignments/{}'.format(self.base_url, self.course_id)
        print(url)
        response = {'success': True, 'assignments': []}
        self.requests_mocker.get(url, json=response)

        url = '{}/assignment/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id
        )
        self.requests_mocker.post(url, json=self._post_assignment)

    def _mock_requests_released(self):
        '''
        Mocks ngshare's GET assignments, which responds with the assignment,
        and POST assignment, which responds with "Assignment already exists".
        '''
        url = '{}/assignments/{}'.format(self.base_url, self.course_id)
        response = {'success': True, 'assignments': [self.assignment_id]}
        self.requests_mocker.get(url, json=response)

        url = '{}/assignment/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id
        )
        response = {'success': False, 'message': 'Assignment already exists'}
        self.requests_mocker.post(url, json=response)

    def _mock_requests_force_rerelease(self):
        '''
        Mocks ngshare's GET assignments, which responds with the assignment if
        DELETE assignment has not been called or no assignment otherwise, POST
        assignment, which responds with "Assignment already exists" if DELETE
        assignment has not been called or success otherwise, and DELETE
        assignment, which notes that the assignment no longer exists.
        '''
        url = '{}/assignments/{}'.format(self.base_url, self.course_id)
        self.requests_mocker.get(url, json=self._get_assignments)

        url = '{}/assignment/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id
        )
        self.requests_mocker.post(url, json=self._post_assignment)

        url = '{}/assignment/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id
        )
        self.requests_mocker.delete(url, json=self._delete_assignment)

    def _new_release_assignment(
        self,
        course_id=TestExchange.course_id,
        assignment_id=TestExchange.assignment_id,
        student_id=TestExchange.student_id,
    ):
        return self._new_exchange_object(
            ExchangeReleaseAssignment, course_id, assignment_id, student_id
        )

    def _post_assignment(self, request: PreparedRequest, context):
        if self.released:
            return {'success': False, 'message': 'Assignment already exists'}

        request = parse_body(request.body)
        try:
            files = loads(request['files'])
            assert len(files) == 1
            notebook_name = self.notebook_id + '.ipynb'
            assert files[0]['path'] == notebook_name
            actual_content = base64.b64decode(files[0]['content'].encode())
            reference_file = self.files_path / 'test.ipynb'
            with open(reference_file, 'rb') as expected_content:
                assert actual_content == expected_content.read()
        except Exception as e:
            self.test_failed = True
            getLogger().error(e)
        self.test_completed = True
        return {'success': True}

    def _prepare_assignment(self):
        assignment_dir = self.course_dir / 'release' / self.assignment_id
        files_dir = Path(__file__).parent / 'files'
        os.makedirs(assignment_dir)
        shutil.copyfile(
            files_dir / 'test.ipynb',
            assignment_dir / (self.notebook_id + '.ipynb'),
        )

    @pytest.fixture(autouse=True)
    def init_release_assignment(self):
        self._prepare_assignment()
        self.release_assignment = self._new_release_assignment()

    def test_404(self):
        self.mock_404()
        try:
            self.release_assignment.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_unsuccessful(self):
        self.mock_unsuccessful()
        try:
            self.release_assignment.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_no_course_id(self, tmpdir_factory):
        """Does releasing without a course id thrown an error?"""
        self.release_assignment.coursedir.course_id = ''
        with pytest.raises(ExchangeError):
            self.release_assignment.start()

    def test_release(self):
        self.released = False
        self._mock_requests_release()
        self.release_assignment.start()

        assert not self.test_failed
        assert self.test_completed

    def test_rerelease(self):
        self._mock_requests_released()
        with pytest.raises(ExchangeError):
            self.release_assignment.start()

    def test_force_rerelease(self):
        self.released = True
        self._mock_requests_force_rerelease()
        self.release_assignment.force = True
        self.release_assignment.start()
        assert not self.test_failed
        assert self.test_completed



from .exchange import Exchange
from .fetch_assignment import ExchangeFetchAssignment
from .fetch_feedback import ExchangeFetchFeedback
from .release_assignment import ExchangeReleaseAssignment
from .release_feedback import ExchangeReleaseFeedback
from .list import ExchangeList
from .submit import ExchangeSubmit
from .collect import ExchangeCollect


def configureExchange(c, ngshare_url=None):
    '''Modifies nbgrader configuration to use ngshare_exchange as the exchange.
    c is the nbgrader config you get using get_config.
    To use, simply specify the following in your nbgrader_config.py file:

    import ngshare_exchage
    ngshare_exchange.configureExchange(get_config())
    '''
    c.ExchangeFactory.exchange = Exchange
    c.ExchangeFactory.fetch_assignment = ExchangeFetchAssignment
    c.ExchangeFactory.fetch_feedback = ExchangeFetchFeedback
    c.ExchangeFactory.release_assignment = ExchangeReleaseAssignment
    c.ExchangeFactory.release_feedback = ExchangeReleaseFeedback
    c.ExchangeFactory.list = ExchangeList
    c.ExchangeFactory.submit = ExchangeSubmit
    c.ExchangeFactory.collect = ExchangeCollect
    if ngshare_url:
        c.ExchangeFactory.exchange._ngshare_url = ngshare_url



import base64
from json import loads
from logging import getLogger
from pathlib import Path
import os
import shutil

import pytest
from requests import PreparedRequest

from .base import parse_body, TestExchange
from nbgrader.exchange import ExchangeError
from .. import ExchangeReleaseFeedback


class TestExchangeReleaseFeedback(TestExchange):

    feedback_file = ''
    timestamp = ''

    def _mock_requests_release(self):
        url = '{}/feedback/{}/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id, self.student_id
        )
        self.requests_mocker.post(url, json=self._post_feedback)

    def _mock_assignment_not_found(self):
        url = '{}/feedback/{}/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id, self.student_id
        )
        response = {'success': False, 'message': 'Assignment not found'}
        self.requests_mocker.post(url, json=response)

    def _post_feedback(self, request: PreparedRequest, context):
        request = parse_body(request.body)
        try:
            timestamp = request['timestamp']
            assert timestamp == self.timestamp
            files = loads(request['files'])
            assert len(files) == 1
            feedback_name = self.notebook_id + '.html'
            assert files[0]['path'] == feedback_name
            actual_content = base64.b64decode(files[0]['content'].encode())
            reference_file = self.files_path / self.feedback_file
            with open(reference_file, 'rb') as expected_content:
                assert actual_content == expected_content.read()
        except Exception as e:
            self.test_failed = True
            getLogger().error(e)
        self.test_completed = True
        return {'success': True}

    def _new_release_feedback(
        self,
        course_id=TestExchange.course_id,
        assignment_id=TestExchange.assignment_id,
        student_id=TestExchange.student_id,
    ):
        return self._new_exchange_object(
            ExchangeReleaseFeedback, course_id, assignment_id, student_id
        )

    def _prepare_feedback(self):
        feedback_dir = (
            self.course_dir / 'feedback' / self.student_id / self.assignment_id
        )
        files_dir = Path(__file__).parent / 'files'
        os.makedirs(feedback_dir)
        shutil.copyfile(
            files_dir / 'feedback.html',
            feedback_dir / (self.notebook_id + '.html'),
        )
        with open(feedback_dir / 'timestamp.txt', 'w') as f:
            f.write('some_timestamp')

    def _prepare_feedback_2(self):
        feedback_dir = (
            self.course_dir / 'feedback' / self.student_id / self.assignment_id
        )
        files_dir = Path(__file__).parent / 'files'
        shutil.copyfile(
            files_dir / 'feedback-changed.html',
            feedback_dir / (self.notebook_id + '.html'),
        )
        with open(feedback_dir / 'timestamp.txt', 'w') as f:
            f.write('some_other_timestamp')

    @pytest.fixture(autouse=True)
    def init_release_feedback(self):
        self._prepare_feedback()
        self.release_feedback = self._new_release_feedback()

    def test_404(self):
        self.mock_404()
        try:
            self.release_feedback.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_unsuccessful(self):
        self.mock_unsuccessful()
        try:
            self.release_feedback.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_no_course_id(self, tmpdir_factory):
        """Does releasing without a course id thrown an error?"""

        self.release_feedback.coursedir.course_id = ''
        with pytest.raises(ExchangeError):
            self.release_feedback.start()

    def test_release(self):
        self.feedback_file = 'feedback.html'
        self.timestamp = 'some_timestamp'
        self._mock_requests_release()
        self.release_feedback.start()
        assert not self.test_failed
        assert self.test_completed

    def test_release_assignment_not_found(self):
        self._mock_assignment_not_found()
        with pytest.raises(ExchangeError):
            self.release_feedback.start()

    def test_rerelease(self):
        self.feedback_file = 'feedback.html'
        self.timestamp = 'some_timestamp'
        self._mock_requests_release()
        self.release_feedback.start()
        assert not self.test_failed
        assert self.test_completed

        self.feedback_file = 'feedback-changed.html'
        self.timestamp = 'some_other_timestamp'
        self._prepare_feedback_2()
        self._mock_requests_release()
        self.release_feedback.start()
        assert not self.test_failed
        assert self.test_completed

    def test_release_multiple_students(self):
        feedback_dir1 = (
            self.course_dir / 'feedback' / 'student_2' / self.assignment_id
        )
        feedback_dir2 = (
            self.course_dir / 'feedback' / 'student_1' / self.assignment_id
        )
        self.student_id = 'student_2'
        self._prepare_feedback()
        assert os.path.exists(feedback_dir1)
        assert os.path.exists(feedback_dir2)
        self.student_id = 'student_1'
        self._mock_requests_release()
        self.student_id = 'student_2'
        self._mock_requests_release()
        self.release_feedback.start()



import base64
import hashlib
from pathlib import Path
import shutil

import pytest

from .base import TestExchange
from nbgrader.auth import Authenticator
from nbgrader.exchange import ExchangeError
from .. import ExchangeFetchAssignment


class TestExchangeFetchAssignment(TestExchange):
    def _mock_requests_fetch(self):
        '''
        Mock's ngshare's GET assignment, which responds with the assignment.
        '''
        url = '{}/assignment/{}/{}'.format(
            self.base_url, self.course_id, self.assignment_id
        )
        content = None
        with open(self.files_path / 'test.ipynb', 'rb') as notebook:
            content = notebook.read()
        md5 = hashlib.md5()
        md5.update(content)
        checksum = md5.hexdigest()
        content = base64.b64encode(content).decode()
        files = [
            {
                'path': self.notebook_id + '.ipynb',
                'content': content,
                'checksum': checksum,
            }
        ]
        response = {'success': True, 'files': files}
        self.requests_mocker.get(url, json=response)

    def _new_fetch_assignment(
        self,
        course_id=TestExchange.course_id,
        assignment_id=TestExchange.assignment_id,
        student_id=TestExchange.student_id,
    ):
        retvalue = self._new_exchange_object(
            ExchangeFetchAssignment, course_id, assignment_id, student_id
        )

        class DummyAuthenticator(Authenticator):
            def has_access(self, student_id, course_id):
                return True

        retvalue.authenticator = DummyAuthenticator()
        retvalue.assignment_dir = str(self.course_dir.absolute())
        return retvalue

    @pytest.fixture(autouse=True)
    def init_fetch_assignment(self):
        self.fetch_assignment = self._new_fetch_assignment()
        self._mock_requests_fetch()

    def test_404(self):
        self.mock_404()
        try:
            self.fetch_assignment.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_unsuccessful(self):
        self.mock_unsuccessful()
        try:
            self.fetch_assignment.start()
        except Exception as e:
            assert issubclass(type(e), ExchangeError)

    def test_no_course_id(self):
        self.fetch_assignment.coursedir.course_id = ''
        with pytest.raises(ExchangeError):
            self.fetch_assignment.start()

    def test_fetch(self):
        self.fetch_assignment.start()
        notebook_path = (
            self.course_dir / self.assignment_id / (self.notebook_id + '.ipynb')
        )
        assert notebook_path.is_file()
        with open(self.files_path / 'test.ipynb', 'rb') as reference_file, open(
            notebook_path, 'rb'
        ) as actual_file:
            assert actual_file.read() == reference_file.read()

    def test_refetch(self):
        self.fetch_assignment.start()
        with pytest.raises(ExchangeError):
            self.fetch_assignment.start()

    def test_replace(self):
        self.fetch_assignment.start()
        self.fetch_assignment.replace_missing_files = True
        self.fetch_assignment.start()

    def test_replace_no_overwrite(self):
        self.fetch_assignment.start()
        self.fetch_assignment.replace_missing_files = True
        # Make sure files aren't overwritten.
        notebook_path = (
            self.course_dir / self.assignment_id / (self.notebook_id + '.ipynb')
        )
        shutil.copyfile(
            self.files_path / 'submitted-changed.ipynb', notebook_path
        )
        with open(notebook_path, 'rb') as file:
            contents1 = file.read()
        self.fetch_assignment.start()
        with open(notebook_path, 'rb') as file:
            contents2 = file.read()
        assert contents1 == contents2

    def test_fetch_multiple_courses(self, tmpdir_factory):
        self.fetch_assignment.start()
        notebook_path = (
            self.course_dir / self.assignment_id / (self.notebook_id + '.ipynb')
        )
        assert notebook_path.is_file()

        self.course_id = 'abc102'
        self.course_dir = Path(tmpdir_factory.mktemp(self.course_id))
        self.fetch_assignment = self._new_fetch_assignment(
            course_id=self.course_id
        )
        self._mock_requests_fetch()
        self.fetch_assignment.start()
        notebook_path = (
            self.course_dir / self.assignment_id / (self.notebook_id + '.ipynb')
        )
        assert notebook_path.is_file()



#!/usr/bin/python
import os
import shutil

from traitlets import Bool

from nbgrader.exchange.abc import (
    ExchangeFetchAssignment as ABCExchangeFetchAssignment,
)
from .exchange import Exchange
from nbgrader.utils import check_mode


class ExchangeFetchAssignment(Exchange, ABCExchangeFetchAssignment):
    def _load_config(self, cfg, **kwargs):
        if 'ExchangeFetch' in cfg:
            self.log.warning(
                'Use ExchangeFetchAssignment in config, not ExchangeFetch. Outdated config:\n%s',
                '\n'.join(
                    'ExchangeFetch.{key} = {value!r}'.format(
                        key=key, value=value
                    )
                    for (key, value) in cfg.ExchangeFetchAssignment.items()
                ),
            )

            cfg.ExchangeFetchAssignment.merge(cfg.ExchangeFetch)
            del cfg.ExchangeFetchAssignment

        super(ExchangeFetchAssignment, self)._load_config(cfg, **kwargs)

    def init_src(self):
        if self.coursedir.course_id == '':
            self.fail('No course id specified. Re-run with --course flag.')
        if not self.authenticator.has_access(
            self.coursedir.student_id, self.coursedir.course_id
        ):
            self.fail('You do not have access to this course.')

        self.src_path = '/assignment/{}/{}'.format(
            self.coursedir.course_id, self.coursedir.assignment_id
        )

    def init_dest(self):
        if self.path_includes_course:
            root = os.path.join(
                self.coursedir.course_id, self.coursedir.assignment_id
            )
        else:
            root = self.coursedir.assignment_id
        self.dest_path = os.path.abspath(
            os.path.join(self.assignment_dir, root)
        )
        if os.path.isdir(self.dest_path) and not self.replace_missing_files:
            self.fail(
                'You already have a copy of the assignment in this directory: {}'.format(
                    root
                )
            )

    def do_copy(self, files):
        """Copy the src dir to the dest dir omitting the self.coursedir.ignore globs."""
        if os.path.isdir(self.dest_path):
            self.coursedir.ignore = True
            self.decode_dir(
                files, self.dest_path, ignore=self.ignore_patterns()
            )
        else:
            self.decode_dir(files, self.dest_path)

        self.log.info(
            'Successfully decoded {}.'.format(self.coursedir.assignment_id)
        )

    def copy_files(self):
        response = self.ngshare_api_get(self.src_path)
        if response is None:
            self.log.warning('Failed to fetch assignment.')
        else:
            self.log.info(
                'Successfully fetched {}. Will try to decode'.format(
                    self.coursedir.assignment_id
                )
            )
            try:
                self.do_copy(response['files'])
            except:
                self.log.warning('Could not decode the assignment')



import os
import glob
import re
import base64
import json

from nbgrader.exchange.abc import (
    ExchangeReleaseFeedback as ABCExchangeReleaseFeedback,
)
from .exchange import Exchange


class ExchangeReleaseFeedback(Exchange, ABCExchangeReleaseFeedback):
    def init_src(self):
        student_id = (
            self.coursedir.student_id if self.coursedir.student_id else '*'
        )
        self.src_path = self.coursedir.format_path(
            self.coursedir.feedback_directory,
            student_id,
            self.coursedir.assignment_id,
        )

    def init_dest(self):
        if self.coursedir.course_id == '':
            self.fail("No course id specified. Re-run with --course flag.")

    def copy_files(self):
        if self.coursedir.student_id_exclude:
            exclude_students = set(self.coursedir.student_id_exclude.split(','))
        else:
            exclude_students = set()

        staged_feedback = {}  # Maps student IDs to submissions.
        html_files = glob.glob(os.path.join(self.src_path, "*.html"))
        for html_file in html_files:
            regexp = re.escape(os.path.sep).join(
                [
                    self.coursedir.format_path(
                        self.coursedir.feedback_directory,
                        "(?P<student_id>.*)",
                        self.coursedir.assignment_id,
                        escape=True,
                    ),
                    "(?P<notebook_id>.*).html",
                ]
            )

            m = re.match(regexp, html_file)
            if m is None:
                msg = "Could not match '%s' with regexp '%s'" % (
                    html_file,
                    regexp,
                )
                self.log.error(msg)
                continue

            gd = m.groupdict()
            student_id = gd['student_id']
            notebook_id = gd['notebook_id']
            if student_id in exclude_students:
                self.log.debug("Skipping student '{}'".format(student_id))
                continue

            feedback_dir = os.path.split(html_file)[0]
            with open(
                os.path.join(feedback_dir, 'timestamp.txt')
            ) as timestamp_file:
                timestamp = timestamp_file.read()

            if student_id not in staged_feedback.keys():
                staged_feedback[student_id] = {}  # Maps timestamp to feedback.
            if timestamp not in staged_feedback[student_id].keys():
                staged_feedback[student_id][timestamp] = []  # List of info.
            staged_feedback[student_id][timestamp].append(
                {'notebook_id': notebook_id, 'path': html_file}
            )

        for student_id, submission in staged_feedback.items():  # Student.
            for timestamp, feedback_info in submission.items():  # Submission.
                self.log.info(
                    "Releasing feedback for student '{}' on "
                    "assignment '{}/{}/{}' ({})".format(
                        student_id,
                        self.coursedir.course_id,
                        self.coursedir.assignment_id,
                        notebook_id,
                        timestamp,
                    )
                )
                retvalue = self.post_feedback(
                    student_id, timestamp, feedback_info
                )
                if retvalue is None:
                    self.fail('Failed to upload feedback to server.')
                else:
                    self.log.info('Feedback released.')

    def post_feedback(self, student_id, timestamp, feedback_info):
        """
        Uploads feedback files for a specific submission.
        ``feedback_info`` - A list of feedback files. Each feedback file is
        represented as a dictionary with a "path" to the local feedback file and
        "notebook_id" of the corresponding notebook.
        """
        url = '/feedback/{}/{}/{}'.format(
            self.coursedir.course_id, self.coursedir.assignment_id, student_id
        )
        files = json.dumps(
            [
                self.encode_file(x['path'], '{}.html'.format(x['notebook_id']))
                for x in feedback_info
            ]
        )
        data = {'timestamp': timestamp, 'files': files}

        return self.ngshare_api_post(url, data)

    # TODO: Consider moving into Exchange.
    def encode_file(self, filesystem_path, assignment_path):
        with open(filesystem_path, 'rb') as f:
            content = f.read()
        return {
            'path': assignment_path,
            'content': base64.encodebytes(content).decode(),
        }



#!/usr/bin/python
import os
import shutil
import glob
import requests
import fnmatch

from textwrap import dedent

from traitlets import Unicode, Bool, default
from jupyter_core.paths import jupyter_data_dir

from nbgrader.exchange.abc import Exchange as ABCExchange
from nbgrader.exchange import ExchangeError
from nbgrader.utils import ignore_patterns
import base64
import json


class Exchange(ABCExchange):

    username = (
        os.environ['JUPYTERHUB_USER']
        if 'JUPYTERHUB_USER' in os.environ
        else os.environ['USER']
    )

    _ngshare_url = Unicode(
        help=dedent(
            """
            Override default ngshare URL
            """
        ),
    ).tag(config=True)

    @property
    def ngshare_url(self):
        if self._ngshare_url:
            return self._ngshare_url
        if 'PROXY_PUBLIC_SERVICE_HOST' in os.environ:
            # we are in a kubernetes environment, so dns based service discovery should work
            # assuming the service is called ngshare, which it should
            return "http://proxy-public/services/ngshare"
        else:
            raise ValueError(
                "ngshare url not configured in a non-k8s environment! Please configure the URL manually in nbgrader_config.py"
            )

    def _ngshare_api_check_error(self, response, url):
        if response.status_code != requests.codes.ok:
            self.log.error(
                "ngshare service returned invalid status code %d.",
                response.status_code,
            )

        try:
            response = response.json()
        except Exception:
            self.log.exception(
                "ngshare service returned non-JSON content: '%s'.",
                response.text,
            )
            return None

        if not response['success']:
            if 'message' not in response:
                self.log.error(
                    "ngshare endpoint %s returned failure without an error message.",
                    url,
                )
            else:
                self.log.error(
                    "ngshare endpoint %s returned failure: %s",
                    url,
                    response['message'],
                )
            return None
        return response

    def ngshare_api_request(self, method, url, data=None, params=None):
        try:
            headers = None
            if 'JUPYTERHUB_API_TOKEN' in os.environ:
                headers = {
                    'Authorization': 'token '
                    + os.environ['JUPYTERHUB_API_TOKEN']
                }
            response = requests.request(
                method,
                self.ngshare_url + url,
                headers=headers,
                data=data,
                params=params,
            )
        except Exception:
            self.log.exception(
                'An error occurred when querying the ngshare ' 'endpoint %s',
                url,
            )
            return None
        return self._ngshare_api_check_error(response, url)

    def ngshare_api_get(self, url, params=None):
        return self.ngshare_api_request('GET', url, params=params)

    def ngshare_api_post(self, url, data, params=None):
        return self.ngshare_api_request('POST', url, data=data, params=params)

    def ngshare_api_delete(self, url, params=None):
        return self.ngshare_api_request('DELETE', url, params=params)

    assignment_dir = Unicode(
        '.',
        help=dedent(
            """
            Local path for storing student assignments.  Defaults to '.'
            which is normally Jupyter's notebook_dir.
            """
        ),
    ).tag(config=True)

    cache = Unicode(
        '',
        help='Local cache directory for nbgrader submit and nbgrader list. Defaults to $JUPYTER_DATA_DIR/nbgrader_cache',
    ).tag(config=True)

    @default('cache')
    def _cache_default(self):
        return os.path.join(jupyter_data_dir(), 'nbgrader_cache')

    path_includes_course = Bool(
        False,
        help=dedent(
            """
            Whether the path for fetching/submitting  assignments should be
            prefixed with the course name. If this is `False`, then the path
            will be something like `./ps1`. If this is `True`, then the path
            will be something like `./course123/ps1`.
            """
        ),
    ).tag(config=True)

    def decode_dir(self, src_dir, dest_dir, ignore=None):
        '''
       decode an encoded directory tree and saw the decoded files to des
       src_dir: en encoded directory tree
       dest: destination directory path for decoded files
       ignore: a function that returns true if the file should be ignored,
       false otherwise. This function takes as arguments the file directory path,
       file name, and file size in KB.
       '''
        # check if the destination directory exists
        if not os.path.exists(dest_dir):
            os.mkdir(dest_dir)

        for src_file in src_dir:
            src_path = src_file['path']
            path_components = os.path.split(src_path)
            dir_name = path_components[0]
            file_name = path_components[1]

            dest_path = os.path.join(dest_dir, file_name)
            # the file could be in a subdirectory, check if directory exists
            if not os.path.exists(dir_name) and dir_name != '':
                subdir = os.path.join(dest_dir, dir_name)
                if not os.path.exists(subdir):
                    os.mkdir(subdir)
                dest_path = os.path.join(subdir, file_name)

            self.log.info('Decoding: {}'.format(dest_path))
            decoded_content = base64.b64decode(src_file['content'])
            file_size = len(decoded_content)

            if ignore:
                if ignore(dest_path, file_name, file_size):
                    continue

            with open(dest_path, 'wb') as d:
                d.write(decoded_content)

    def encode_dir(self, src_dir, ignore=None):
        encoded_files = []
        for (subdir, dirs, files) in os.walk(src_dir):
            for file_name in files:
                file_path = subdir + os.sep + file_name
                data_bytes = open(file_path, 'rb').read()
                if ignore:
                    if ignore(subdir, file_name, len(data_bytes)):
                        continue

                # check if you have a subdir
                sub_dir = subdir.split(os.sep)[-1]
                if sub_dir != self.coursedir.assignment_id:
                    file_path = sub_dir + os.sep + file_name
                else:
                    file_path = file_name

                self.log.info('Encoding: {}'.format(file_path))
                encoded = base64.b64encode(data_bytes)
                content = str(encoded, 'utf-8')
                file_map = {'path': file_path, 'content': content}
                encoded_files.append(file_map)

        dir_tree = {'user': self.username, 'files': json.dumps(encoded_files)}
        return dir_tree

    def init_src(self):
        """Compute and check the source paths for the transfer."""

        raise NotImplementedError

    def init_dest(self):
        """Compute and check the destination paths for the transfer."""

        raise NotImplementedError

    def copy_files(self):
        """Actually do the file transfer."""

        raise NotImplementedError

    def start(self):
        return super(Exchange, self).start()

    def _assignment_not_found(self, src_path, other_path):
        msg = 'Assignment not found at: {}'.format(src_path)
        self.log.fatal(msg)
        found = glob.glob(other_path)
        if found:

            # Normally it is a bad idea to put imports in the middle of
            # a function, but we do this here because otherwise fuzzywuzzy
            # prints an annoying message about python-Levenshtein every
            # time nbgrader is run.

            from fuzzywuzzy import fuzz

            scores = sorted([(fuzz.ratio(self.src_path, x), x) for x in found])
            self.log.error('Did you mean: %s', scores[-1][1])

        raise ExchangeError(msg)

    def do_copy(self, src, dest, log=None):
        """
        Copy the src dir to the dest dir, omitting excluded
        file/directories, non included files, and too large files, as
        specified by the options coursedir.ignore, coursedir.include
        and coursedir.max_file_size.
        """
        shutil.copytree(
            src,
            dest,
            ignore=ignore_patterns(
                exclude=self.coursedir.ignore,
                include=self.coursedir.include,
                max_file_size=self.coursedir.max_file_size,
                log=self.log,
            ),
        )
        # copytree copies access mode too - so we must add go+rw back to it if
        # we are in groupshared.
        if self.coursedir.groupshared:
            for dirname, _, filenames in os.walk(dest):
                # dirs become ug+rwx
                st_mode = os.stat(dirname).st_mode
                if st_mode & 0o2770 != 0o2770:
                    try:
                        os.chmod(dirname, (st_mode | 0o2770) & 0o2777)
                    except PermissionError:
                        self.log.warning(
                            "Could not update permissions of %s to make it groupshared",
                            dirname,
                        )

                for filename in filenames:
                    filename = os.path.join(dirname, filename)
                    st_mode = os.stat(filename).st_mode
                    if st_mode & 0o660 != 0o660:
                        try:
                            os.chmod(filename, (st_mode | 0o660) & 0o777)
                        except PermissionError:
                            self.log.warning(
                                "Could not update permissions of %s to make it groupshared",
                                filename,
                            )

    def ignore_patterns(self):
        """
        Returns a function which decides whether or not a file should be
        ignored. The function has the signature
            ignore_patterns(directory, filename, filesize) -> bool
        The directory and filename are the same parameters as described in
        shutil.ignore_patterns. filesize is the size of the file in kilobytes.
        All filenames matching patterns in self.coursedir.ignore, if it exists,
        will be ignored. If self.coursedir.include exists, filenames not
        matching the patterns will be ignored. If self.coursedir.max_file_size
        exists, files exceeding that size in kilobytes will be ignored.
        """
        exclude = self.coursedir.ignore
        include = self.coursedir.include
        max_file_size = self.coursedir.max_file_size
        log = self.log

        def ignore_patterns(directory, filename, filesize):
            fullname = os.path.join(directory, filename)
            if exclude and any(
                fnmatch.fnmatch(filename, glob) for glob in exclude
            ):
                if log:
                    log.debug(
                        "Ignoring excluded file '{}' (see config option "
                        'CourseDirectory.ignore)'.format(fullname)
                    )
                return True
            elif include and not any(
                fnmatch.fnmatch(filename, glob) for glob in include
            ):
                if log:
                    log.debug(
                        "Ignoring non included file '{}' (see config "
                        'option CourseDirectory.include)'.format(fullname)
                    )
                return True
            elif max_file_size and filesize > 1000 * max_file_size:
                if log:
                    log.warning(
                        "Ignoring file too large '{}' (see config "
                        'option CourseDirectory.max_file_size)'.format(fullname)
                    )
                return True
            return False

        return ignore_patterns
